#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Base HDF-EOS reader."""

from __future__ import annotations

import logging
import re
from ast import literal_eval
from contextlib import suppress
from datetime import datetime

import numpy as np
import xarray as xr
from pyhdf.error import HDF4Error
from pyhdf.SD import SD

from satpy import CHUNK_SIZE, DataID
from satpy.readers.file_handlers import BaseFileHandler

logger = logging.getLogger(__name__)


def interpolate(clons, clats, csatz, src_resolution, dst_resolution):
    """Interpolate two parallel datasets jointly."""
    if csatz is None:
        return _interpolate_no_angles(clons, clats, src_resolution, dst_resolution)
    return _interpolate_with_angles(clons, clats, csatz, src_resolution, dst_resolution)


def _interpolate_with_angles(clons, clats, csatz, src_resolution, dst_resolution):
    from geotiepoints.modisinterpolator import modis_1km_to_250m, modis_1km_to_500m, modis_5km_to_1km

    # (src_res, dst_res, is satz not None) -> interp function
    interpolation_functions = {
        (5000, 1000): modis_5km_to_1km,
        (1000, 500): modis_1km_to_500m,
        (1000, 250): modis_1km_to_250m
    }
    return _find_and_run_interpolation(interpolation_functions, src_resolution, dst_resolution,
                                       (clons, clats, csatz))


def _interpolate_no_angles(clons, clats, src_resolution, dst_resolution):
    interpolation_functions = {}

    try:
        from geotiepoints.simple_modis_interpolator import modis_1km_to_250m as simple_1km_to_250m
        from geotiepoints.simple_modis_interpolator import modis_1km_to_500m as simple_1km_to_500m
    except ImportError:
        raise NotImplementedError(
            f"Interpolation from {src_resolution}m to {dst_resolution}m "
            "without satellite zenith angle information is not "
            "implemented. Try updating your version of "
            "python-geotiepoints.")
    else:
        interpolation_functions[(1000, 500)] = simple_1km_to_500m
        interpolation_functions[(1000, 250)] = simple_1km_to_250m

    return _find_and_run_interpolation(interpolation_functions, src_resolution, dst_resolution,
                                       (clons, clats))


def _find_and_run_interpolation(interpolation_functions, src_resolution, dst_resolution, args):
    try:
        interpolation_function = interpolation_functions[(src_resolution, dst_resolution)]
    except KeyError:
        error_message = "Interpolation from {}m to {}m not implemented".format(
            src_resolution, dst_resolution)
        raise NotImplementedError(error_message)

    logger.debug("Interpolating from {} to {}".format(src_resolution, dst_resolution))
    return interpolation_function(*args)


class HDFEOSBaseFileReader(BaseFileHandler):
    """Base file handler for HDF EOS data for both L1b and L2 products."""

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize the base reader."""
        BaseFileHandler.__init__(self, filename, filename_info, filetype_info)
        try:
            self.sd = SD(self.filename)
        except HDF4Error as err:
            error_message = "Could not load data from file {}: {}".format(self.filename, err)
            raise ValueError(error_message)

        self.metadata = self._load_all_metadata_attributes()

    def _load_all_metadata_attributes(self):
        metadata = {}
        attrs = self.sd.attributes()
        for md_key in ("CoreMetadata.0", "StructMetadata.0", "ArchiveMetadata.0"):
            try:
                str_val = attrs[md_key]
            except KeyError:
                continue
            else:
                metadata.update(self.read_mda(str_val))
        return metadata

    @classmethod
    def read_mda(cls, attribute):
        """Read the EOS metadata."""
        line_iterator = iter(attribute.split('\n'))
        return cls._read_mda(line_iterator)

    @classmethod
    def _read_mda(cls, lines, element=None):
        current_dict = {}

        for line in lines:
            if not line:
                continue
            if line == 'END':
                return current_dict

            key, val = cls._split_line(line, lines)

            if key in ['GROUP', 'OBJECT']:
                current_dict[val] = cls._read_mda(lines, val)
            elif key in ['END_GROUP', 'END_OBJECT']:
                if val != element:
                    raise SyntaxError("Non-matching end-tag")
                return current_dict
            elif key in ['CLASS', 'NUM_VAL']:
                pass
            else:
                current_dict[key] = val
        logger.warning("Malformed EOS metadata, missing an END.")
        return current_dict

    @classmethod
    def _split_line(cls, line, lines):
        key, val = line.split('=')
        key = key.strip()
        val = val.strip()
        try:
            with suppress(ValueError):
                val = literal_eval(val)
        except SyntaxError:
            key, val = cls._split_line(line + next(lines), lines)
        return key, val

    @property
    def metadata_platform_name(self):
        """Platform name from the internal file metadata."""
        try:
            # Example: 'Terra' or 'Aqua'
            return self.metadata['INVENTORYMETADATA']['ASSOCIATEDPLATFORMINSTRUMENTSENSOR'][
                'ASSOCIATEDPLATFORMINSTRUMENTSENSORCONTAINER']['ASSOCIATEDPLATFORMSHORTNAME']['VALUE']
        except KeyError:
            return self._platform_name_from_filename()

    def _platform_name_from_filename(self):
        platform_indicator = self.filename_info["platform_indicator"]
        if platform_indicator in ("t", "O"):
            # t1.* or MOD*
            return "Terra"
        # a1.* or MYD*
        return "Aqua"

    @property
    def start_time(self):
        """Get the start time of the dataset."""
        try:
            date = (self.metadata['INVENTORYMETADATA']['RANGEDATETIME']['RANGEBEGINNINGDATE']['VALUE'] + ' ' +
                    self.metadata['INVENTORYMETADATA']['RANGEDATETIME']['RANGEBEGINNINGTIME']['VALUE'])
            return datetime.strptime(date, '%Y-%m-%d %H:%M:%S.%f')
        except KeyError:
            return self._start_time_from_filename()

    def _start_time_from_filename(self):
        for fn_key in ("start_time", "acquisition_time"):
            if fn_key in self.filename_info:
                return self.filename_info[fn_key]
        raise RuntimeError("Could not determine file start time")

    @property
    def end_time(self):
        """Get the end time of the dataset."""
        try:
            date = (self.metadata['INVENTORYMETADATA']['RANGEDATETIME']['RANGEENDINGDATE']['VALUE'] + ' ' +
                    self.metadata['INVENTORYMETADATA']['RANGEDATETIME']['RANGEENDINGTIME']['VALUE'])
            return datetime.strptime(date, '%Y-%m-%d %H:%M:%S.%f')
        except KeyError:
            return self.start_time

    def _read_dataset_in_file(self, dataset_name):
        if dataset_name not in self.sd.datasets():
            error_message = "Dataset name {} not included in available datasets {}".format(
                dataset_name, self.sd.datasets()
            )
            raise KeyError(error_message)

        dataset = self.sd.select(dataset_name)
        return dataset

    def load_dataset(self, dataset_name, is_category=False):
        """Load the dataset from HDF EOS file."""
        from satpy.readers.hdf4_utils import from_sds

        dataset = self._read_dataset_in_file(dataset_name)
        dask_arr = from_sds(dataset, chunks=CHUNK_SIZE)
        dims = ('y', 'x') if dask_arr.ndim == 2 else None
        data = xr.DataArray(dask_arr, dims=dims,
                            attrs=dataset.attributes())
        data = self._scale_and_mask_data_array(data, is_category=is_category)

        return data

    def _scale_and_mask_data_array(self, data, is_category=False):
        good_mask, new_fill = self._get_good_data_mask(data, is_category=is_category)
        scale_factor = data.attrs.pop('scale_factor', None)
        add_offset = data.attrs.pop('add_offset', None)
        # don't scale category products, even though scale_factor may equal 1
        # we still need to convert integers to floats
        if scale_factor is not None and not is_category:
            data = data * np.float32(scale_factor)
            if add_offset is not None and add_offset != 0:
                data = data + add_offset

        if good_mask is not None:
            data = data.where(good_mask, new_fill)
        return data

    def _get_good_data_mask(self, data_arr, is_category=False):
        try:
            fill_value = data_arr.attrs["_FillValue"]
        except KeyError:
            return None, None

        # preserve integer data types if possible
        if is_category and np.issubdtype(data_arr.dtype, np.integer):
            # no need to mask, the fill value is already what it needs to be
            return None, None
        new_fill = np.nan
        data_arr.attrs.pop('_FillValue', None)
        good_mask = data_arr != fill_value
        return good_mask, new_fill

    def _add_satpy_metadata(self, data_id: DataID, data_arr: xr.DataArray):
        """Add metadata that is specific to Satpy."""
        new_attrs = {
            'platform_name': 'EOS-' + self.metadata_platform_name,
            'sensor': 'modis',
        }

        res = data_id["resolution"]
        rps = self._resolution_to_rows_per_scan(res)
        new_attrs["rows_per_scan"] = rps

        data_arr.attrs.update(new_attrs)

    def _resolution_to_rows_per_scan(self, resolution: int) -> int:
        known_rps = {
            5000: 2,
            1000: 10,
            500: 20,
            250: 40,
        }
        return known_rps.get(resolution, 10)


class HDFEOSGeoReader(HDFEOSBaseFileReader):
    """Handler for the geographical datasets."""

    # list of geographical datasets handled by the georeader
    # mapping to the default variable name if not specified in YAML
    DATASET_NAMES = {
        'longitude': 'Longitude',
        'latitude': 'Latitude',
        'satellite_azimuth_angle': ('SensorAzimuth', 'Sensor_Azimuth'),
        'satellite_zenith_angle': ('SensorZenith', 'Sensor_Zenith'),
        'solar_azimuth_angle': ('SolarAzimuth', 'SolarAzimuth'),
        'solar_zenith_angle': ('SolarZenith', 'Solar_Zenith'),
    }

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize the geographical reader."""
        HDFEOSBaseFileReader.__init__(self, filename, filename_info, filetype_info)
        self.cache = {}

    @staticmethod
    def is_geo_loadable_dataset(dataset_name: str) -> bool:
        """Determine if this dataset should be loaded as a Geo dataset."""
        return dataset_name in HDFEOSGeoReader.DATASET_NAMES

    @staticmethod
    def read_geo_resolution(metadata):
        """Parse metadata to find the geolocation resolution."""
        # level 1 files
        try:
            return HDFEOSGeoReader._geo_resolution_for_l1b(metadata)
        except KeyError:
            try:
                return HDFEOSGeoReader._geo_resolution_for_l2_l1b(metadata)
            except (AttributeError, KeyError):
                raise RuntimeError("Could not determine resolution from file metadata")

    @staticmethod
    def _geo_resolution_for_l1b(metadata):
        ds = metadata['INVENTORYMETADATA']['COLLECTIONDESCRIPTIONCLASS']['SHORTNAME']['VALUE']
        if ds.endswith('D03') or ds.endswith('HKM') or ds.endswith('QKM'):
            return 1000
        # 1km files have 5km geolocation usually
        return 5000

    @staticmethod
    def _geo_resolution_for_l2_l1b(metadata):
        # data files probably have this level 2 files
        # this does not work for L1B 1KM data files because they are listed
        # as 1KM data but the geo data inside is at 5km
        latitude_dim = metadata['SwathStructure']['SWATH_1']['DimensionMap']['DimensionMap_2']['GeoDimension']
        resolution_regex = re.compile(r'(?P<resolution>\d+)(km|KM)')
        resolution_match = resolution_regex.search(latitude_dim)
        return int(resolution_match.group('resolution')) * 1000

    @property
    def geo_resolution(self):
        """Resolution of the geographical data retrieved in the metadata."""
        return self.read_geo_resolution(self.metadata)

    def _load_ds_by_name(self, ds_name):
        """Attempt loading using multiple common names."""
        var_names = self.DATASET_NAMES[ds_name]
        if isinstance(var_names, (list, tuple)):
            try:
                return self.load_dataset(var_names[0])
            except KeyError:
                return self.load_dataset(var_names[1])
        return self.load_dataset(var_names)

    def get_interpolated_dataset(self, name1, name2, resolution, offset=0):
        """Load and interpolate datasets."""
        try:
            result1 = self.cache[(name1, resolution)]
            result2 = self.cache[(name2, resolution)]
        except KeyError:
            result1 = self._load_ds_by_name(name1)
            result2 = self._load_ds_by_name(name2) - offset
            try:
                sensor_zenith = self._load_ds_by_name('satellite_zenith_angle')
            except KeyError:
                # no sensor zenith angle, do "simple" interpolation
                sensor_zenith = None

            result1, result2 = interpolate(
                result1, result2, sensor_zenith,
                self.geo_resolution, resolution
            )
            self.cache[(name1, resolution)] = result1
            self.cache[(name2, resolution)] = result2 + offset

    def get_dataset(self, dataset_id: DataID, dataset_info: dict) -> xr.DataArray:
        """Get the geolocation dataset."""
        # Name of the dataset as it appears in the HDF EOS file
        in_file_dataset_name = dataset_info.get('file_key')
        # Name of the dataset in the YAML file
        dataset_name = dataset_id['name']
        # Resolution asked
        resolution = dataset_id['resolution']
        if in_file_dataset_name is not None:
            # if the YAML was configured with a specific name use that
            data = self.load_dataset(in_file_dataset_name)
        else:
            # otherwise use the default name for this variable
            data = self._load_ds_by_name(dataset_name)
        if resolution != self.geo_resolution:
            if in_file_dataset_name is not None:
                # they specified a custom variable name but
                # we don't know how to interpolate this yet
                raise NotImplementedError(
                    "Interpolation for variable '{}' is not "
                    "configured".format(dataset_name))

            # The data must be interpolated
            logger.debug("Loading %s", dataset_name)
            if dataset_name in ['longitude', 'latitude']:
                self.get_interpolated_dataset('longitude', 'latitude',
                                              resolution)
            elif dataset_name in ['satellite_azimuth_angle', 'satellite_zenith_angle']:
                # Sensor dataset names differs between L1b and L2 products
                self.get_interpolated_dataset('satellite_azimuth_angle', 'satellite_zenith_angle',
                                              resolution, offset=90)
            elif dataset_name in ['solar_azimuth_angle', 'solar_zenith_angle']:
                # Sensor dataset names differs between L1b and L2 products
                self.get_interpolated_dataset('solar_azimuth_angle', 'solar_zenith_angle',
                                              resolution, offset=90)

            data = self.cache[dataset_name, resolution]

        for key in ('standard_name', 'units'):
            if key in dataset_info:
                data.attrs[key] = dataset_info[key]
        self._add_satpy_metadata(dataset_id, data)

        return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Satpy developers
#
# This file is part of Satpy.
#
# Satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.tropomi_l2 module."""

import os
from datetime import datetime
from unittest import mock

import numpy as np
import pytest
import xarray as xr

METOP_FILE = "IMG_SX.M2.D17037.S1601.E1607.B0000001.WE.HR.ORB.nc"
NPP_MIRS_L2_SWATH = "NPR-MIRS-IMG_v11r6_npp_s201702061601000_e201702061607000_c202012201658410.nc"
N20_MIRS_L2_SWATH = "NPR-MIRS-IMG_v11r4_n20_s201702061601000_e201702061607000_c202012201658410.nc"
OTHER_MIRS_L2_SWATH = "NPR-MIRS-IMG_v11r4_gpm_s201702061601000_e201702061607000_c202010080001310.nc"

EXAMPLE_FILES = [METOP_FILE, NPP_MIRS_L2_SWATH, OTHER_MIRS_L2_SWATH]

N_CHANNEL = 22
N_FOV = 96
N_SCANLINE = 100
DEFAULT_FILE_DTYPE = np.float64
DEFAULT_2D_SHAPE = (N_SCANLINE, N_FOV)
DEFAULT_DATE = datetime(2019, 6, 19, 13, 0)
DEFAULT_LAT = np.linspace(23.09356, 36.42844, N_SCANLINE * N_FOV,
                          dtype=DEFAULT_FILE_DTYPE)
DEFAULT_LON = np.linspace(127.6879, 144.5284, N_SCANLINE * N_FOV,
                          dtype=DEFAULT_FILE_DTYPE)
FREQ = xr.DataArray([23.8, 31.4, 50.3, 51.76, 52.8, 53.596, 54.4, 54.94, 55.5,
                     57.29, 57.29, 57.29, 57.29, 57.29, 57.29, 88.2, 165.5,
                     183.31, 183.31, 183.31, 183.31, 183.31][:N_CHANNEL],
                    dims='Channel',
                    attrs={'description': "Central Frequencies (GHz)"})
POLO = xr.DataArray([2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 3, 3, 3,
                     3, 3, 3][:N_CHANNEL], dims='Channel',
                    attrs={'description': "Polarizations"})

DS_IDS = ['RR', 'longitude', 'latitude']
TEST_VARS = ['btemp_88v', 'btemp_165h',
             'btemp_23v', 'RR', 'Sfc_type']
DEFAULT_UNITS = {'btemp_88v': 'K', 'btemp_165h': 'K',
                 'btemp_23v': 'K', 'RR': 'mm/hr', 'Sfc_type': "1"}
PLATFORM = {"M2": "metop-a", "NPP": "npp", "GPM": "gpm"}
SENSOR = {"m2": "amsu-mhs", "npp": "atms", "gpm": "GPI"}

START_TIME = datetime(2017, 2, 6, 16, 1, 0)
END_TIME = datetime(2017, 2, 6, 16, 7, 0)


def fake_coeff_from_fn(fn):
    """Create Fake Coefficients."""
    ameans = np.random.uniform(261, 267, N_CHANNEL)
    locations = [
        [1, 2],
        [1, 2],
        [3, 4, 5],
        [3, 4, 5],
        [4, 5, 6],
        [5, 6, 7],
        [6, 7, 8],
        [7, 8],
        [9, 10, 11],
        [10, 11],
        [10, 11, 12],
        [11, 12, 13],
        [12, 13],
        [12, 13, 14],
        [14, 15],
        [1, 16],
        [17, 18],
        [18, 19],
        [18, 19, 20],
        [19, 20, 21],
        [20, 21, 22],
        [21, 22],
    ]
    all_nchx = [len(loc) for loc in locations]

    coeff_str = []
    for idx in range(1, N_CHANNEL + 1):
        nx = idx - 1
        coeff_str.append('\n')
        next_line = '   {}  {} {}\n'.format(idx, all_nchx[nx], ameans[nx])
        coeff_str.append(next_line)
        next_line = '   {}\n'.format("   ".join([str(x) for x in locations[idx - 1]]))
        coeff_str.append(next_line)
        for fov in range(1, N_FOV+1):
            random_coeff = np.random.rand(all_nchx[nx])
            random_coeff = np.ones(all_nchx[nx])
            str_coeff = '  '.join([str(x) for x in random_coeff])
            random_means = np.random.uniform(261, 267, all_nchx[nx])
            random_means = np.zeros(all_nchx[nx])
            str_means = ' '.join([str(x) for x in random_means])
            error_val = np.random.uniform(0, 4)
            coeffs_line = ' {:>2} {:>2}  {} {}   {}\n'.format(idx, fov,
                                                              str_coeff,
                                                              str_means,
                                                              error_val)
            coeff_str.append(coeffs_line)

    return coeff_str


def _get_datasets_with_attributes(**kwargs):
    """Represent files with two resolution of variables in them (ex. OCEAN)."""
    bt = xr.DataArray(np.linspace(1830, 3930, N_SCANLINE * N_FOV * N_CHANNEL).
                      reshape(N_SCANLINE, N_FOV, N_CHANNEL),
                      attrs={'long_name': "Channel Temperature (K)",
                             'units': "Kelvin",
                             'coordinates': "Longitude Latitude Freq",
                             'scale_factor': 0.01,
                             '_FillValue': -999,
                             'valid_range': [0, 50000]},
                      dims=('Scanline', 'Field_of_view', 'Channel'))
    rr = xr.DataArray(np.random.randint(100, 500, size=(N_SCANLINE, N_FOV)),
                      attrs={'long_name': "Rain Rate (mm/hr)",
                             'units': "mm/hr",
                             'coordinates': "Longitude Latitude",
                             'scale_factor': 0.1,
                             '_FillValue': -999,
                             'valid_range': [0, 1000]},
                      dims=('Scanline', 'Field_of_view'))
    sfc_type = xr.DataArray(np.random.randint(0, 4, size=(N_SCANLINE, N_FOV)),
                            attrs={'description': "type of surface:0-ocean," +
                                                  "1-sea ice,2-land,3-snow",
                                   'units': "1",
                                   'coordinates': "Longitude Latitude",
                                   '_FillValue': -999,
                                   'valid_range': [0, 3]
                                   },
                            dims=('Scanline', 'Field_of_view'))
    latitude = xr.DataArray(DEFAULT_LAT.reshape(DEFAULT_2D_SHAPE),
                            attrs={'long_name':
                                   "Latitude of the view (-90,90)"},
                            dims=('Scanline', 'Field_of_view'))
    longitude = xr.DataArray(DEFAULT_LON.reshape(DEFAULT_2D_SHAPE),
                             attrs={'long_name':
                                    "Longitude of the view (-180,180)"},
                             dims=('Scanline', 'Field_of_view'))

    ds_vars = {
        'Freq': FREQ,
        'Polo': POLO,
        'BT': bt,
        'RR': rr,
        'Sfc_type': sfc_type,
        'Latitude': latitude,
        'Longitude': longitude
    }

    attrs = {'missing_value': -999.}
    ds = xr.Dataset(ds_vars, attrs=attrs)
    ds = ds.assign_coords({"Freq": FREQ, "Latitude": latitude, "Longitude": longitude})
    return ds


def _get_datasets_with_less_attributes():
    """Represent files with two resolution of variables in them (ex. OCEAN)."""
    bt = xr.DataArray(np.linspace(1830, 3930, N_SCANLINE * N_FOV * N_CHANNEL).
                      reshape(N_SCANLINE, N_FOV, N_CHANNEL),
                      attrs={'long_name': "Channel Temperature (K)",
                             'scale_factor': 0.01},
                      dims=('Scanline', 'Field_of_view', 'Channel'))
    rr = xr.DataArray(np.random.randint(100, 500, size=(N_SCANLINE, N_FOV)),
                      attrs={'long_name': "Rain Rate (mm/hr)",
                             'scale_factor': 0.1},
                      dims=('Scanline', 'Field_of_view'))

    sfc_type = xr.DataArray(np.random.randint(0, 4, size=(N_SCANLINE, N_FOV)),
                            attrs={'description': "type of surface:0-ocean," +
                                                  "1-sea ice,2-land,3-snow"},
                            dims=('Scanline', 'Field_of_view'))
    latitude = xr.DataArray(DEFAULT_LAT.reshape(DEFAULT_2D_SHAPE),
                            attrs={'long_name':
                                   "Latitude of the view (-90,90)"},
                            dims=('Scanline', 'Field_of_view'))
    longitude = xr.DataArray(DEFAULT_LON.reshape(DEFAULT_2D_SHAPE),
                             attrs={"long_name":
                                    "Longitude of the view (-180,180)"},
                             dims=('Scanline', 'Field_of_view'))

    ds_vars = {
        'Freq': FREQ,
        'Polo': POLO,
        'BT': bt,
        'RR': rr,
        'Sfc_type': sfc_type,
        'Longitude': longitude,
        'Latitude': latitude
    }

    attrs = {'missing_value': -999.}
    ds = xr.Dataset(ds_vars, attrs=attrs)
    ds = ds.assign_coords({"Freq": FREQ, "Latitude": latitude, "Longitude": longitude})
    return ds


def fake_open_dataset(filename, **kwargs):
    """Create a Dataset similar to reading an actual file with xarray.open_dataset."""
    if filename == METOP_FILE:
        return _get_datasets_with_less_attributes()
    return _get_datasets_with_attributes()


class TestMirsL2_NcReader:
    """Test mirs Reader."""

    yaml_file = "mirs.yaml"

    def setup_method(self):
        """Read fake data."""
        from satpy._config import config_search_paths
        self.reader_configs = config_search_paths(os.path.join('readers', self.yaml_file))

    @pytest.mark.parametrize(
        ("filenames", "expected_loadables"),
        [
            ([METOP_FILE], 1),
            ([NPP_MIRS_L2_SWATH], 1),
            ([OTHER_MIRS_L2_SWATH], 1),
        ]
    )
    def test_reader_creation(self, filenames, expected_loadables):
        """Test basic initialization."""
        from satpy.readers import load_reader
        with mock.patch('satpy.readers.mirs.xr.open_dataset') as od:
            od.side_effect = fake_open_dataset
            r = load_reader(self.reader_configs)
            loadables = r.select_files_from_pathnames(filenames)
            assert len(loadables) == expected_loadables
            r.create_filehandlers(loadables)
            # make sure we have some files
            assert r.file_handlers

    @pytest.mark.parametrize(
        ("filenames", "expected_datasets"),
        [
            ([METOP_FILE], DS_IDS),
            ([NPP_MIRS_L2_SWATH], DS_IDS),
            ([OTHER_MIRS_L2_SWATH], DS_IDS),
        ]
    )
    def test_available_datasets(self, filenames, expected_datasets):
        """Test that variables are dynamically discovered."""
        from satpy.readers import load_reader
        with mock.patch('satpy.readers.mirs.xr.open_dataset') as od:
            od.side_effect = fake_open_dataset
            r = load_reader(self.reader_configs)
            loadables = r.select_files_from_pathnames(filenames)
            r.create_filehandlers(loadables)
            avails = list(r.available_dataset_names)
            for var_name in expected_datasets:
                assert var_name in avails

    @staticmethod
    def _check_area(data_arr):
        from pyresample.geometry import SwathDefinition
        area = data_arr.attrs['area']
        assert isinstance(area, SwathDefinition)

    @staticmethod
    def _check_fill(data_arr):
        assert '_FillValue' not in data_arr.attrs
        if np.issubdtype(data_arr.dtype, np.floating):
            # we started with float32, it should stay that way
            assert data_arr.dtype.type == np.float64

    @staticmethod
    def _check_valid_range(data_arr, test_valid_range):
        # valid_range is popped out of data_arr.attrs when it is applied
        assert 'valid_range' not in data_arr.attrs
        assert data_arr.data.min() >= test_valid_range[0]
        assert data_arr.data.max() <= test_valid_range[1]

    @staticmethod
    def _check_fill_value(data_arr, test_fill_value):
        assert '_FillValue' not in data_arr.attrs
        assert not (data_arr.data == test_fill_value).any()

    @staticmethod
    def _check_attrs(data_arr, platform_name):
        attrs = data_arr.attrs
        assert 'scale_factor' not in attrs
        assert 'platform_name' in attrs
        assert attrs['platform_name'] == platform_name
        assert attrs['start_time'] == START_TIME
        assert attrs['end_time'] == END_TIME

    @pytest.mark.parametrize(
        ("filenames", "loadable_ids", "platform_name"),
        [
            ([METOP_FILE], TEST_VARS, "metop-a"),
            ([NPP_MIRS_L2_SWATH], TEST_VARS, "npp"),
            ([N20_MIRS_L2_SWATH], TEST_VARS, "noaa-20"),
            ([OTHER_MIRS_L2_SWATH], TEST_VARS, "gpm"),
        ]
    )
    @pytest.mark.parametrize('reader_kw', [{}, {'limb_correction': False}])
    def test_basic_load(self, filenames, loadable_ids,
                        platform_name, reader_kw):
        """Test that variables are loaded properly."""
        from satpy.readers import load_reader
        with mock.patch('satpy.readers.mirs.xr.open_dataset') as od:
            od.side_effect = fake_open_dataset
            r = load_reader(self.reader_configs)
            loadables = r.select_files_from_pathnames(filenames)
            r.create_filehandlers(loadables,  fh_kwargs=reader_kw)
            with mock.patch('satpy.readers.mirs.read_atms_coeff_to_string') as \
                    fd, mock.patch('satpy.readers.mirs.retrieve'):
                fd.side_effect = fake_coeff_from_fn
                loaded_data_arrs = r.load(loadable_ids)
            assert len(loaded_data_arrs) == len(loadable_ids)

            test_data = fake_open_dataset(filenames[0])
            for _data_id, data_arr in loaded_data_arrs.items():
                data_arr = data_arr.compute()
                var_name = data_arr.attrs["name"]
                if var_name not in ['latitude', 'longitude']:
                    self._check_area(data_arr)
                self._check_fill(data_arr)
                self._check_attrs(data_arr, platform_name)

                input_fake_data = test_data['BT'] if "btemp" in var_name \
                    else test_data[var_name]
                if "valid_range" in input_fake_data.attrs:
                    valid_range = input_fake_data.attrs["valid_range"]
                    self._check_valid_range(data_arr, valid_range)
                if "_FillValue" in input_fake_data.attrs:
                    fill_value = input_fake_data.attrs["_FillValue"]
                    self._check_fill_value(data_arr, fill_value)

                sensor = data_arr.attrs['sensor']
                if reader_kw.get('limb_correction', True) and sensor == 'atms':
                    fd.assert_called()
                else:
                    fd.assert_not_called()
                assert data_arr.attrs['units'] == DEFAULT_UNITS[var_name]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2012-2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Reader for aapp level 1b data.

Options for loading:

 - pre_launch_coeffs (False): use pre-launch coefficients if True, operational
   otherwise (if available).

https://nwp-saf.eumetsat.int/site/download/documentation/aapp/NWPSAF-MF-UD-003_Formats_v8.0.pdf
"""
import functools
import logging
from datetime import datetime, timedelta

import dask.array as da
import numpy as np
import xarray as xr
from dask import delayed

from satpy import CHUNK_SIZE
from satpy.readers.file_handlers import BaseFileHandler

LINE_CHUNK = CHUNK_SIZE ** 2 // 2048

logger = logging.getLogger(__name__)

AVHRR_CHANNEL_NAMES = ["1", "2", "3a", "3b", "4", "5"]

AVHRR_ANGLE_NAMES = ['sensor_zenith_angle',
                     'solar_zenith_angle',
                     'sun_sensor_azimuth_difference_angle']

AVHRR_PLATFORM_IDS2NAMES = {4: 'NOAA-15',
                            2: 'NOAA-16',
                            6: 'NOAA-17',
                            7: 'NOAA-18',
                            8: 'NOAA-19',
                            11: 'Metop-B',
                            12: 'Metop-A',
                            13: 'Metop-C',
                            14: 'Metop simulator'}


def create_xarray(arr):
    """Create an `xarray.DataArray`."""
    res = xr.DataArray(arr, dims=['y', 'x'])
    return res


class AAPPL1BaseFileHandler(BaseFileHandler):
    """A base file handler for the AAPP level-1 formats."""

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize AAPP level-1 file handler object."""
        super().__init__(filename, filename_info, filetype_info)

        self.channels = None
        self.units = None
        self.sensor = "unknown"

        self._data = None
        self._header = None
        self.area = None

        self._channel_names = []
        self._angle_names = []

    def _set_filedata_layout(self):
        """Set the file data type/layout."""
        self._header_offset = 0
        self._scan_type = np.dtype([("siteid", "<i2")])
        self._header_type = np.dtype([("siteid", "<i2")])

    @property
    def start_time(self):
        """Get the time of the first observation."""
        return datetime(self._data['scnlinyr'][0], 1, 1) + timedelta(
            days=int(self._data['scnlindy'][0]) - 1,
            milliseconds=int(self._data['scnlintime'][0]))

    @property
    def end_time(self):
        """Get the time of the final observation."""
        return datetime(self._data['scnlinyr'][-1], 1, 1) + timedelta(
            days=int(self._data['scnlindy'][-1]) - 1,
            milliseconds=int(self._data['scnlintime'][-1]))

    def _update_dataset_attributes(self, dataset, key, info):
        dataset.attrs.update({'platform_name': self.platform_name,
                              'sensor': self.sensor})
        dataset.attrs.update(key.to_dict())
        for meta_key in ('standard_name', 'units'):
            if meta_key in info:
                dataset.attrs.setdefault(meta_key, info[meta_key])

    def _get_platform_name(self, platform_names_lookup):
        """Get the platform name from the file header."""
        self.platform_name = platform_names_lookup.get(self._header['satid'][0], None)
        if self.platform_name is None:
            raise ValueError("Unsupported platform ID: %d" % self.header['satid'])

    def read(self):
        """Read the data."""
        tic = datetime.now()
        header = np.memmap(self.filename, dtype=self._header_type, mode="r", shape=(1, ))
        data = np.memmap(self.filename, dtype=self._scan_type, offset=self._header_offset, mode="r")
        logger.debug("Reading time %s", str(datetime.now() - tic))

        self._header = header
        self._data = data

    def _calibrate_active_channel_data(self, key):
        """Calibrate active channel data only."""
        raise NotImplementedError("This should be implemented in the sub class")

    def get_dataset(self, key, info):
        """Get a dataset from the file."""
        if key['name'] in self._channel_names:
            dataset = self._calibrate_active_channel_data(key)
            if dataset is None:
                return None
        elif key['name'] in ['longitude', 'latitude']:
            dataset = self.navigate(key['name'])
            dataset.attrs = info
        elif key['name'] in self._angle_names:
            dataset = self.get_angles(key['name'])
        else:
            raise ValueError("Not a supported dataset: %s", key['name'])

        self._update_dataset_attributes(dataset, key, info)
        return dataset


class AVHRRAAPPL1BFile(AAPPL1BaseFileHandler):
    """Reader for AVHRR L1B files created from the AAPP software."""

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize object information by reading the input file."""
        super(AVHRRAAPPL1BFile, self).__init__(filename, filename_info,
                                               filetype_info)

        self.channels = {i: None for i in AVHRR_CHANNEL_NAMES}
        self.units = {i: 'counts' for i in AVHRR_CHANNEL_NAMES}

        self._is3b = None
        self._is3a = None
        self._channel_names = AVHRR_CHANNEL_NAMES
        self._angle_names = AVHRR_ANGLE_NAMES

        self._set_filedata_layout()
        self.read()

        self.active_channels = self._get_active_channels()

        self._get_platform_name(AVHRR_PLATFORM_IDS2NAMES)
        self.sensor = 'avhrr-3'

    def _set_filedata_layout(self):
        """Set the file data type/layout."""
        self._header_offset = 22016
        self._scan_type = _SCANTYPE
        self._header_type = _HEADERTYPE

    def _get_active_channels(self):
        status = self._get_channel_binary_status_from_header()
        return self._convert_binary_channel_status_to_activation_dict(status)

    def _calibrate_active_channel_data(self, key):
        """Calibrate active channel data only."""
        if self.active_channels[key['name']]:
            return self.calibrate(key)
        return None

    def _get_channel_binary_status_from_header(self):
        status = self._header['inststat1'].item()
        change_line = self._header['statchrecnb']
        if change_line > 0:
            status |= self._header['inststat2'].item()
        return status

    @staticmethod
    def _convert_binary_channel_status_to_activation_dict(status):
        bits_channels = ((13, '1'),
                         (12, '2'),
                         (11, '3a'),
                         (10, '3b'),
                         (9, '4'),
                         (8, '5'))
        activated = dict()
        for bit, channel_name in bits_channels:
            activated[channel_name] = bool(status >> bit & 1)
        return activated

    def available_datasets(self, configured_datasets=None):
        """Get the available datasets."""
        for _, mda in configured_datasets:
            if mda['name'] in self._channel_names:
                yield self.active_channels[mda['name']], mda
            else:
                yield True, mda

    def get_angles(self, angle_id):
        """Get sun-satellite viewing angles."""
        sunz, satz, azidiff = self._get_all_interpolated_angles()

        name_to_variable = dict(zip(self._angle_names, (satz, sunz, azidiff)))
        return create_xarray(name_to_variable[angle_id])

    @functools.lru_cache(maxsize=10)
    def _get_all_interpolated_angles(self):
        sunz40km, satz40km, azidiff40km = self._get_tiepoint_angles_in_degrees()
        return self._interpolate_arrays(sunz40km, satz40km, azidiff40km)

    def _get_tiepoint_angles_in_degrees(self):
        sunz40km = self._data["ang"][:, :, 0] * 1e-2
        satz40km = self._data["ang"][:, :, 1] * 1e-2
        azidiff40km = self._data["ang"][:, :, 2] * 1e-2
        return sunz40km, satz40km, azidiff40km

    def _interpolate_arrays(self, *input_arrays, geolocation=False):
        lines = input_arrays[0].shape[0]
        try:
            interpolator = self._create_40km_interpolator(lines, *input_arrays, geolocation=geolocation)
        except ImportError:
            logger.warning("Could not interpolate, python-geotiepoints missing.")
            output_arrays = input_arrays
        else:
            output_delayed = delayed(interpolator.interpolate, nout=3)()
            output_arrays = [da.from_delayed(out_array, (lines, 2048), in_array.dtype)
                             for in_array, out_array in zip(input_arrays, output_delayed)]
        return output_arrays

    @staticmethod
    def _create_40km_interpolator(lines, *arrays_40km, geolocation=False):
        if geolocation:
            # Slower but accurate at datum line
            from geotiepoints.geointerpolator import GeoInterpolator as Interpolator
        else:
            from geotiepoints.interpolator import Interpolator
        cols40km = np.arange(24, 2048, 40)
        cols1km = np.arange(2048)
        rows40km = np.arange(lines)
        rows1km = np.arange(lines)
        along_track_order = 1
        cross_track_order = 3
        satint = Interpolator(
            arrays_40km, (rows40km, cols40km),
            (rows1km, cols1km), along_track_order, cross_track_order)
        return satint

    def navigate(self, coordinate_id):
        """Get the longitudes and latitudes of the scene."""
        lons, lats = self._get_all_interpolated_coordinates()
        if coordinate_id == 'longitude':
            return create_xarray(lons)
        if coordinate_id == 'latitude':
            return create_xarray(lats)

        raise KeyError("Coordinate {} unknown.".format(coordinate_id))

    @functools.lru_cache(maxsize=10)
    def _get_all_interpolated_coordinates(self):
        lons40km, lats40km = self._get_coordinates_in_degrees()
        return self._interpolate_arrays(lons40km, lats40km, geolocation=True)

    def _get_coordinates_in_degrees(self):
        lons40km = self._data["pos"][:, :, 1] * 1e-4
        lats40km = self._data["pos"][:, :, 0] * 1e-4
        return lons40km, lats40km

    def calibrate(self,
                  dataset_id,
                  pre_launch_coeffs=False,
                  calib_coeffs=None):
        """Calibrate the data."""
        if calib_coeffs is None:
            calib_coeffs = {}

        units = {'reflectance': '%',
                 'brightness_temperature': 'K',
                 'counts': '',
                 'radiance': 'W*m-2*sr-1*cm ?'}

        if dataset_id['name'] in ("3a", "3b") and self._is3b is None:
            # Is it 3a or 3b:
            self._is3a = da.bitwise_and(da.from_array(self._data['scnlinbit'],
                                                      chunks=LINE_CHUNK), 3) == 0
            self._is3b = da.bitwise_and(da.from_array(self._data['scnlinbit'],
                                                      chunks=LINE_CHUNK), 3) == 1

        try:
            vis_idx = ['1', '2', '3a'].index(dataset_id['name'])
            ir_idx = None
        except ValueError:
            vis_idx = None
            ir_idx = ['3b', '4', '5'].index(dataset_id['name'])

        mask = True
        if vis_idx is not None:
            coeffs = calib_coeffs.get('ch' + dataset_id['name'])
            if dataset_id['name'] == '3a':
                mask = self._is3a[:, None]
            ds = create_xarray(
                _vis_calibrate(self._data,
                               vis_idx,
                               dataset_id['calibration'],
                               pre_launch_coeffs,
                               coeffs,
                               mask=mask))
        else:
            if dataset_id['name'] == '3b':
                mask = self._is3b[:, None]
            ds = create_xarray(
                _ir_calibrate(self._header,
                              self._data,
                              ir_idx,
                              dataset_id['calibration'],
                              mask=mask))

        ds.attrs['units'] = units[dataset_id['calibration']]
        ds.attrs.update(dataset_id._asdict())
        return ds


# AAPP 1b header

_HEADERTYPE = np.dtype([("siteid", "S3"),
                        ("blank", "S1"),
                        ("l1bversnb", "<i2"),
                        ("l1bversyr", "<i2"),
                        ("l1bversdy", "<i2"),
                        ("reclg", "<i2"),
                        ("blksz", "<i2"),
                        ("hdrcnt", "<i2"),
                        ("filler0", "S6"),
                        ("dataname", "S42"),
                        ("prblkid", "S8"),
                        ("satid", "<i2"),
                        ("instid", "<i2"),
                        ("datatype", "<i2"),
                        ("tipsrc", "<i2"),
                        ("startdatajd", "<i4"),
                        ("startdatayr", "<i2"),
                        ("startdatady", "<i2"),
                        ("startdatatime", "<i4"),
                        ("enddatajd", "<i4"),
                        ("enddatayr", "<i2"),
                        ("enddatady", "<i2"),
                        ("enddatatime", "<i4"),
                        ("cpidsyr", "<i2"),
                        ("cpidsdy", "<i2"),
                        ("filler1", "S8"),
                        # data set quality indicators
                        ("inststat1", "<i4"),
                        ("filler2", "S2"),
                        ("statchrecnb", "<i2"),
                        ("inststat2", "<i4"),
                        ("scnlin", "<i2"),
                        ("callocscnlin", "<i2"),
                        ("misscnlin", "<i2"),
                        ("datagaps", "<i2"),
                        ("okdatafr", "<i2"),
                        ("pacsparityerr", "<i2"),
                        ("auxsyncerrsum", "<i2"),
                        ("timeseqerr", "<i2"),
                        ("timeseqerrcode", "<i2"),
                        ("socclockupind", "<i2"),
                        ("locerrind", "<i2"),
                        ("locerrcode", "<i2"),
                        ("pacsstatfield", "<i2"),
                        ("pacsdatasrc", "<i2"),
                        ("filler3", "S4"),
                        ("spare1", "S8"),
                        ("spare2", "S8"),
                        ("filler4", "S10"),
                        # Calibration
                        ("racalind", "<i2"),
                        ("solarcalyr", "<i2"),
                        ("solarcaldy", "<i2"),
                        ("pcalalgind", "<i2"),
                        ("pcalalgopt", "<i2"),
                        ("scalalgind", "<i2"),
                        ("scalalgopt", "<i2"),
                        ("irttcoef", "<i2", (4, 6)),
                        ("filler5", "<i4", (2, )),
                        # radiance to temperature conversion
                        ("albcnv", "<i4", (2, 3)),
                        ("radtempcnv", "<i4", (3, 3)),
                        ("filler6", "<i4", (3, )),
                        # Navigation
                        ("modelid", "S8"),
                        ("nadloctol", "<i2"),
                        ("locbit", "<i2"),
                        ("filler7", "S2"),
                        ("rollerr", "<i2"),
                        ("pitcherr", "<i2"),
                        ("yawerr", "<i2"),
                        ("epoyr", "<i2"),
                        ("epody", "<i2"),
                        ("epotime", "<i4"),
                        ("smaxis", "<i4"),
                        ("eccen", "<i4"),
                        ("incli", "<i4"),
                        ("argper", "<i4"),
                        ("rascnod", "<i4"),
                        ("manom", "<i4"),
                        ("xpos", "<i4"),
                        ("ypos", "<i4"),
                        ("zpos", "<i4"),
                        ("xvel", "<i4"),
                        ("yvel", "<i4"),
                        ("zvel", "<i4"),
                        ("earthsun", "<i4"),
                        ("filler8", "S16"),
                        # analog telemetry conversion
                        ("pchtemp", "<i2", (5, )),
                        ("reserved1", "<i2"),
                        ("pchtempext", "<i2", (5, )),
                        ("reserved2", "<i2"),
                        ("pchpow", "<i2", (5, )),
                        ("reserved3", "<i2"),
                        ("rdtemp", "<i2", (5, )),
                        ("reserved4", "<i2"),
                        ("bbtemp1", "<i2", (5, )),
                        ("reserved5", "<i2"),
                        ("bbtemp2", "<i2", (5, )),
                        ("reserved6", "<i2"),
                        ("bbtemp3", "<i2", (5, )),
                        ("reserved7", "<i2"),
                        ("bbtemp4", "<i2", (5, )),
                        ("reserved8", "<i2"),
                        ("eleccur", "<i2", (5, )),
                        ("reserved9", "<i2"),
                        ("motorcur", "<i2", (5, )),
                        ("reserved10", "<i2"),
                        ("earthpos", "<i2", (5, )),
                        ("reserved11", "<i2"),
                        ("electemp", "<i2", (5, )),
                        ("reserved12", "<i2"),
                        ("chtemp", "<i2", (5, )),
                        ("reserved13", "<i2"),
                        ("bptemp", "<i2", (5, )),
                        ("reserved14", "<i2"),
                        ("mhtemp", "<i2", (5, )),
                        ("reserved15", "<i2"),
                        ("adcontemp", "<i2", (5, )),
                        ("reserved16", "<i2"),
                        ("d4bvolt", "<i2", (5, )),
                        ("reserved17", "<i2"),
                        ("d5bvolt", "<i2", (5, )),
                        ("reserved18", "<i2"),
                        ("bbtempchn3b", "<i2", (5, )),
                        ("reserved19", "<i2"),
                        ("bbtempchn4", "<i2", (5, )),
                        ("reserved20", "<i2"),
                        ("bbtempchn5", "<i2", (5, )),
                        ("reserved21", "<i2"),
                        ("refvolt", "<i2", (5, )),
                        ("reserved22", "<i2"), ])

# AAPP 1b scanline

_SCANTYPE = np.dtype([("scnlin", "<i2"),
                      ("scnlinyr", "<i2"),
                      ("scnlindy", "<i2"),
                      ("clockdrift", "<i2"),
                      ("scnlintime", "<i4"),
                      ("scnlinbit", "<i2"),
                      ("filler0", "S10"),
                      ("qualind", "<i4"),
                      ("scnlinqual", "<i4"),
                      ("calqual", "<i2", (3, )),
                      ("cbiterr", "<i2"),
                      ("filler1", "S8"),
                      # Calibration
                      ("calvis", "<i4", (3, 3, 5)),
                      ("calir", "<i4", (3, 2, 3)),
                      ("filler2", "<i4", (3, )),
                      # Navigation
                      ("navstat", "<i4"),
                      ("attangtime", "<i4"),
                      ("rollang", "<i2"),
                      ("pitchang", "<i2"),
                      ("yawang", "<i2"),
                      ("scalti", "<i2"),
                      ("ang", "<i2", (51, 3)),
                      ("filler3", "<i2", (3, )),
                      ("pos", "<i4", (51, 2)),
                      ("filler4", "<i4", (2, )),
                      ("telem", "<i2", (103, )),
                      ("filler5", "<i2"),
                      ("hrpt", "<i2", (2048, 5)),
                      ("filler6", "<i4", (2, )),
                      # tip minor frame header
                      ("tipmfhd", "<i2", (7, 5)),
                      # cpu telemetry
                      ("cputel", "S6", (2, 5)),
                      ("filler7", "<i2", (67, )), ])


def _vis_calibrate(data,
                   chn,
                   calib_type,
                   pre_launch_coeffs=False,
                   calib_coeffs=None,
                   mask=True):
    """Calibrate visible channel data.

    *calib_type* in count, reflectance, radiance.

    """
    # Calibration count to albedo, the calibration is performed separately for
    # two value ranges.
    if calib_type not in ['counts', 'radiance', 'reflectance']:
        raise ValueError('Calibration ' + calib_type + ' unknown!')

    channel = da.from_array(data["hrpt"][:, :, chn], chunks=(LINE_CHUNK, 2048))
    mask &= channel != 0

    if calib_type == 'counts':
        return channel

    channel = channel.astype(np.float64)

    if calib_type == 'radiance':
        logger.info("Radiances are not yet supported for " +
                    "the VIS/NIR channels!")

    if pre_launch_coeffs:
        coeff_idx = 2
    else:
        # check that coeffs are valid
        if np.all(data["calvis"][:, chn, 0, 4] == 0):
            logger.info(
                "No valid operational coefficients, fall back to pre-launch")
            coeff_idx = 2
        else:
            coeff_idx = 0

    intersection = da.from_array(data["calvis"][:, chn, coeff_idx, 4],
                                 chunks=LINE_CHUNK)

    if calib_coeffs is not None:
        logger.info("Updating from external calibration coefficients.")
        slope1 = da.from_array(calib_coeffs[0], chunks=LINE_CHUNK)
        intercept1 = da.from_array(calib_coeffs[1], chunks=LINE_CHUNK)
        slope2 = da.from_array(calib_coeffs[2], chunks=LINE_CHUNK)
        intercept2 = da.from_array(calib_coeffs[3], chunks=LINE_CHUNK)
    else:
        slope1 = da.from_array(data["calvis"][:, chn, coeff_idx, 0],
                               chunks=LINE_CHUNK) * 1e-10
        intercept1 = da.from_array(data["calvis"][:, chn, coeff_idx, 1],
                                   chunks=LINE_CHUNK) * 1e-7
        slope2 = da.from_array(data["calvis"][:, chn, coeff_idx, 2],
                               chunks=LINE_CHUNK) * 1e-10
        intercept2 = da.from_array(data["calvis"][:, chn, coeff_idx, 3],
                                   chunks=LINE_CHUNK) * 1e-7

        if chn == 1:
            # In the level 1b file, the visible coefficients are stored as 4-byte integers. Scaling factors then convert
            # them to real numbers which are applied to the measured counts. The coefficient is different depending on
            # whether the counts are less than or greater than the high-gain/low-gain transition value (nominally 500).
            # The slope for visible channels should always be positive (reflectance increases with count). With the
            # pre-launch coefficients the channel 2 slope is always positive but with the operational coefs the stored
            # number in the high-reflectance regime overflows the maximum 2147483647, i.e. it is negative when
            # interpreted as a signed integer. So you have to modify it.
            slope2 = da.where(slope2 < 0, slope2 + 0.4294967296, slope2)

    channel = da.where(channel <= intersection[:, None],
                       channel * slope1[:, None] + intercept1[:, None],
                       channel * slope2[:, None] + intercept2[:, None])

    channel = channel.clip(min=0)

    return da.where(mask, channel, np.nan)


def _ir_calibrate(header, data, irchn, calib_type, mask=True):
    """Calibrate for IR bands.

    *calib_type* in brightness_temperature, radiance, count

    """
    count = da.from_array(data["hrpt"][:, :, irchn + 2], chunks=(LINE_CHUNK, 2048))

    if calib_type == 0:
        return count

    # Mask unnaturally low values
    mask &= count != 0
    count = count.astype(np.float64)

    k1_ = da.from_array(data['calir'][:, irchn, 0, 0], chunks=LINE_CHUNK) / 1.0e9
    k2_ = da.from_array(data['calir'][:, irchn, 0, 1], chunks=LINE_CHUNK) / 1.0e6
    k3_ = da.from_array(data['calir'][:, irchn, 0, 2], chunks=LINE_CHUNK) / 1.0e6

    # Count to radiance conversion:
    rad = k1_[:, None] * count * count + k2_[:, None] * count + k3_[:, None]

    # Suspicious lines
    mask &= ((k1_ != 0) | (k2_ != 0) | (k3_ != 0))[:, None]

    if calib_type == 2:
        mask &= rad > 0.0
        return da.where(mask, rad, np.nan)

    # Central wavenumber:
    cwnum = header['radtempcnv'][0, irchn, 0]
    if irchn == 0:
        cwnum = cwnum / 1.0e2
    else:
        cwnum = cwnum / 1.0e3

    bandcor_2 = header['radtempcnv'][0, irchn, 1] / 1e5
    bandcor_3 = header['radtempcnv'][0, irchn, 2] / 1e6

    ir_const_1 = 1.1910659e-5
    ir_const_2 = 1.438833

    t_planck = (ir_const_2 * cwnum) / \
        np.log(1 + ir_const_1 * cwnum * cwnum * cwnum / rad)

    # Band corrections applied to t_planck to get correct
    # brightness temperature for channel:
    if bandcor_2 < 0:  # Post AAPP-v4
        tb_ = bandcor_2 + bandcor_3 * t_planck
    else:  # AAPP 1 to 4
        tb_ = (t_planck - bandcor_2) / bandcor_3

    # Mask unnaturally low values

    return da.where(mask, tb_, np.nan)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unittesting the Native SEVIRI reader."""

from __future__ import annotations

import os
import unittest
from datetime import datetime
from unittest import mock

import dask.array as da
import numpy as np
import pytest
import xarray as xr

from satpy.readers.eum_base import time_cds_short
from satpy.readers.seviri_l1b_native import ImageBoundaries, NativeMSGFileHandler, Padder, get_available_channels
from satpy.tests.reader_tests.test_seviri_base import ORBIT_POLYNOMIALS, ORBIT_POLYNOMIALS_INVALID
from satpy.tests.reader_tests.test_seviri_l1b_calibration import TestFileHandlerCalibrationBase
from satpy.tests.utils import assert_attrs_equal, make_dataid

CHANNEL_INDEX_LIST = ['VIS006', 'VIS008', 'IR_016', 'IR_039',
                      'WV_062', 'WV_073', 'IR_087', 'IR_097',
                      'IR_108', 'IR_120', 'IR_134', 'HRV']
AVAILABLE_CHANNELS = {}
for item in CHANNEL_INDEX_LIST:
    AVAILABLE_CHANNELS[item] = True

SEC15HDR = '15_SECONDARY_PRODUCT_HEADER'
IDS = 'SelectedBandIDs'

TEST1_HEADER_CHNLIST: dict[str, dict[str, dict]] = {SEC15HDR: {IDS: {}}}
TEST1_HEADER_CHNLIST[SEC15HDR][IDS]['Value'] = 'XX--XX--XX--'

TEST2_HEADER_CHNLIST: dict[str, dict[str, dict]] = {SEC15HDR: {IDS: {}}}
TEST2_HEADER_CHNLIST[SEC15HDR][IDS]['Value'] = 'XX-XXXX----X'

TEST3_HEADER_CHNLIST: dict[str, dict[str, dict]] = {SEC15HDR: {IDS: {}}}
TEST3_HEADER_CHNLIST[SEC15HDR][IDS]['Value'] = 'XXXXXXXXXXXX'

TEST_AREA_EXTENT_EARTHMODEL1_VISIR_FULLDISK = {
    'earth_model': 1,
    'dataset_id': make_dataid(name='VIS006', resolution=3000),
    'is_full_disk': True,
    'is_rapid_scan': 0,
    'fill_disk': False,
    'expected_area_def': {
        'Area ID': 'msg_seviri_fes_3km',
        'Projection': {'a': '6378169000', 'b': '6356583800', 'h': '35785831',
                       'lon_0': '0', 'no_defs': 'None', 'proj': 'geos',
                       'type': 'crs', 'units': 'm', 'x_0': '0', 'y_0': '0'},
        'Number of columns': 3712,
        'Number of rows': 3712,
        'Area extent': (5568748.2758, 5568748.2758, -5568748.2758, -5568748.2758)
    }
}

TEST_AREA_EXTENT_EARTHMODEL1_VISIR_RAPIDSCAN = {
    'earth_model': 1,
    'dataset_id': make_dataid(name='VIS006', resolution=3000),
    'is_full_disk': False,
    'is_rapid_scan': 1,
    'fill_disk': False,
    'expected_area_def': {
        'Area ID': 'msg_seviri_rss_3km',
        'Projection': {'a': '6378169000', 'b': '6356583800', 'h': '35785831',
                       'lon_0': '9.5', 'no_defs': 'None', 'proj': 'geos',
                       'type': 'crs', 'units': 'm', 'x_0': '0', 'y_0': '0'},
        'Number of columns': 3712,
        'Number of rows': 1392,
        'Area extent': (5568748.275756836, 5568748.275756836, -5568748.275756836, 1392187.068939209)
    }
}

TEST_AREA_EXTENT_EARTHMODEL1_VISIR_RAPIDSCAN_FILL = {
    'earth_model': 1,
    'dataset_id': make_dataid(name='VIS006', resolution=3000),
    'is_full_disk': False,
    'is_rapid_scan': 1,
    'fill_disk': True,
    'expected_area_def': {
        'Area ID': 'msg_seviri_rss_3km',
        'Projection': {'a': '6378169000', 'b': '6356583800', 'h': '35785831',
                       'lon_0': '9.5', 'no_defs': 'None', 'proj': 'geos',
                       'type': 'crs', 'units': 'm', 'x_0': '0', 'y_0': '0'},
        'Number of columns': 3712,
        'Number of rows': 3712,
        'Area extent': (5568748.2758, 5568748.2758, -5568748.2758, -5568748.2758)
    }
}

TEST_AREA_EXTENT_EARTHMODEL1_VISIR_ROI = {
    'earth_model': 1,
    'dataset_id': make_dataid(name='VIS006', resolution=3000),
    'is_full_disk': False,
    'is_rapid_scan': 0,
    'fill_disk': False,
    'expected_area_def': {
        'Area ID': 'msg_seviri_fes_3km',
        'Projection': {'a': '6378169000', 'b': '6356583800', 'h': '35785831',
                       'lon_0': '0', 'no_defs': 'None', 'proj': 'geos',
                       'type': 'crs', 'units': 'm', 'x_0': '0', 'y_0': '0'},
        'Number of columns': 2516,
        'Number of rows': 1829,
        'Area extent': (5337717.232, 5154692.6389, -2211297.1332, -333044.7514)

    }
}

TEST_AREA_EXTENT_EARTHMODEL1_VISIR_ROI_FILL = {
    'earth_model': 1,
    'dataset_id': make_dataid(name='VIS006', resolution=3000),
    'is_full_disk': False,
    'is_rapid_scan': 0,
    'fill_disk': True,
    'expected_area_def': {
        'Area ID': 'msg_seviri_fes_3km',
        'Projection': {'a': '6378169000', 'b': '6356583800', 'h': '35785831',
                       'lon_0': '0', 'no_defs': 'None', 'proj': 'geos',
                       'type': 'crs', 'units': 'm', 'x_0': '0', 'y_0': '0'},
        'Number of columns': 3712,
        'Number of rows': 3712,
        'Area extent': (5568748.2758, 5568748.2758, -5568748.2758, -5568748.2758)

    }
}

TEST_AREA_EXTENT_EARTHMODEL1_HRV_FULLDISK = {
    'earth_model': 1,
    'dataset_id': make_dataid(name='HRV', resolution=1000),
    'is_full_disk': True,
    'is_rapid_scan': 0,
    'fill_disk': False,
    'expected_area_def': {
        'Area ID': 'msg_seviri_fes_1km',
        'Projection': {'a': '6378169000', 'b': '6356583800', 'h': '35785831',
                       'lon_0': '0', 'no_defs': 'None', 'proj': 'geos',
                       'type': 'crs', 'units': 'm', 'x_0': '0', 'y_0': '0'},
        'Number of columns': 5568,
        'Number of rows': 11136,
        'Area extent 0': (5567747.920155525, 2625352.665781975, -1000.1343488693237, -5567747.920155525),
        'Area extent 1': (3602483.924627304, 5569748.188853264, -1966264.1298770905, 2625352.665781975)
    }
}

TEST_AREA_EXTENT_EARTHMODEL1_HRV_FULLDISK_FILL = {
    'earth_model': 1,
    'dataset_id': make_dataid(name='HRV', resolution=1000),
    'is_full_disk': True,
    'is_rapid_scan': 0,
    'fill_disk': True,
    'expected_area_def': {
        'Area ID': 'msg_seviri_fes_1km',
        'Projection': {'a': '6378169000', 'b': '6356583800', 'h': '35785831',
                       'lon_0': '0', 'no_defs': 'None', 'proj': 'geos',
                       'type': 'crs', 'units': 'm', 'x_0': '0', 'y_0': '0'},
        'Number of columns': 11136,
        'Number of rows': 11136,
        'Area extent': (5567747.920155525, 5569748.188853264, -5569748.188853264, -5567747.920155525)
    }
}

TEST_AREA_EXTENT_EARTHMODEL1_HRV_RAPIDSCAN = {
    'earth_model': 1,
    'dataset_id': make_dataid(name='HRV', resolution=1000),
    'is_full_disk': False,
    'is_rapid_scan': 1,
    'fill_disk': False,
    'expected_area_def': {
        'Area ID': 'msg_seviri_rss_1km',
        'Projection': {'a': '6378169000', 'b': '6356583800', 'h': '35785831',
                       'lon_0': '9.5', 'no_defs': 'None', 'proj': 'geos',
                       'type': 'crs', 'units': 'm', 'x_0': '0', 'y_0': '0'},
        'Number of columns': 5568,
        'Number of rows': 8192,
        'Area extent': (5567747.920155525, 2625352.665781975, -1000.1343488693237, -5567747.920155525)
    }
}

TEST_AREA_EXTENT_EARTHMODEL1_HRV_RAPIDSCAN_FILL = {
    'earth_model': 1,
    'dataset_id': make_dataid(name='HRV', resolution=1000),
    'is_full_disk': False,
    'is_rapid_scan': 1,
    'fill_disk': True,
    'expected_area_def': {
        'Area ID': 'msg_seviri_rss_1km',
        'Projection': {'a': '6378169000', 'b': '6356583800', 'h': '35785831',
                       'lon_0': '9.5', 'no_defs': 'None', 'proj': 'geos',
                       'type': 'crs', 'units': 'm', 'x_0': '0', 'y_0': '0'},
        'Number of columns': 11136,
        'Number of rows': 11136,
        'Area extent': (5567747.920155525, 5569748.188853264, -5569748.188853264, -5567747.920155525)
    }
}

TEST_AREA_EXTENT_EARTHMODEL1_HRV_ROI = {
    'earth_model': 1,
    'dataset_id': make_dataid(name='HRV', resolution=1000),
    'is_full_disk': False,
    'is_rapid_scan': 0,
    'fill_disk': False,
    'expected_area_def': {
        'Area ID': 'msg_seviri_fes_1km',
        'Projection': {'a': '6378169000', 'b': '6356583800', 'h': '35785831',
                       'lon_0': '0', 'no_defs': 'None', 'proj': 'geos',
                       'type': 'crs', 'units': 'm', 'x_0': '0', 'y_0': '0'},
        'Number of columns': 7548,
        'Number of rows': 5487,
        'Area extent': (5336716.885566711, 5155692.568421364, -2212297.179698944, -332044.6038246155)
    }
}

TEST_AREA_EXTENT_EARTHMODEL1_HRV_ROI_FILL = {
    'earth_model': 1,
    'dataset_id': make_dataid(name='HRV', resolution=1000),
    'is_full_disk': False,
    'is_rapid_scan': 0,
    'fill_disk': True,
    'expected_area_def': {
        'Area ID': 'msg_seviri_fes_1km',
        'Projection': {'a': '6378169000', 'b': '6356583800', 'h': '35785831',
                       'lon_0': '0', 'no_defs': 'None', 'proj': 'geos',
                       'type': 'crs', 'units': 'm', 'x_0': '0', 'y_0': '0'},
        'Number of columns': 11136,
        'Number of rows': 11136,
        'Area extent': (5567747.920155525, 5569748.188853264, -5569748.188853264, -5567747.920155525)
    }
}

TEST_AREA_EXTENT_EARTHMODEL2_VISIR_FULLDISK = {
    'earth_model': 2,
    'dataset_id': make_dataid(name='VIS006', resolution=3000),
    'is_full_disk': True,
    'is_rapid_scan': 0,
    'fill_disk': False,
    'expected_area_def': {
        'Area ID': 'msg_seviri_fes_3km',
        'Projection': {'a': '6378169000', 'b': '6356583800', 'h': '35785831',
                       'lon_0': '0', 'no_defs': 'None', 'proj': 'geos',
                       'type': 'crs', 'units': 'm', 'x_0': '0', 'y_0': '0'},
        'Number of columns': 3712,
        'Number of rows': 3712,
        'Area extent': (5567248.0742, 5570248.4773, -5570248.4773, -5567248.0742)
    }
}

TEST_AREA_EXTENT_EARTHMODEL2_HRV_FULLDISK = {
    'earth_model': 2,
    'dataset_id': make_dataid(name='HRV', resolution=1000),
    'is_full_disk': True,
    'is_rapid_scan': 0,
    'fill_disk': False,
    'expected_area_def': {
        'Area ID': 'msg_seviri_fes_1km',
        'Projection': {'a': '6378169000', 'b': '6356583800', 'h': '35785831',
                       'lon_0': '0', 'no_defs': 'None', 'proj': 'geos',
                       'type': 'crs', 'units': 'm', 'x_0': '0', 'y_0': '0'},
        'Number of columns': 5568,
        'Number of rows': 11136,
        'Area extent 0': (5566247.718632221, 2626852.867305279, -2500.3358721733093, -5566247.718632221),
        'Area extent 1': (3600983.723104, 5571248.390376568, -1967764.3314003944, 2626852.867305279)
    }
}

TEST_AREA_EXTENT_EARTHMODEL2_HRV_FULLDISK_FILL = {
    'earth_model': 2,
    'dataset_id': make_dataid(name='HRV', resolution=1000),
    'is_full_disk': True,
    'is_rapid_scan': 0,
    'fill_disk': True,
    'expected_area_def': {
        'Area ID': 'msg_seviri_fes_1km',
        'Projection': {'a': '6378169000', 'b': '6356583800', 'h': '35785831',
                       'lon_0': '0', 'no_defs': 'None', 'proj': 'geos',
                       'type': 'crs', 'units': 'm', 'x_0': '0', 'y_0': '0'},
        'Number of columns': 11136,
        'Number of rows': 11136,
        'Area extent': (5566247.718632221, 5571248.390376568, -5571248.390376568, -5566247.718632221)
    }
}

TEST_AREA_EXTENT_EARTHMODEL2_VISIR_RAPIDSCAN = {
    'earth_model': 2,
    'dataset_id': make_dataid(name='VIS006', resolution=3000),
    'is_full_disk': False,
    'is_rapid_scan': 1,
    'fill_disk': False,
    'expected_area_def': {
        'Area ID': 'msg_seviri_rss_3km',
        'Projection': {'a': '6378169000', 'b': '6356583800', 'h': '35785831',
                       'lon_0': '9.5', 'no_defs': 'None', 'proj': 'geos',
                       'type': 'crs', 'units': 'm', 'x_0': '0', 'y_0': '0'},
        'Number of columns': 3712,
        'Number of rows': 1392,
        'Area extent': (5567248.074173927, 5570248.477339745, -5570248.477339745, 1393687.2705221176)

    }
}

TEST_AREA_EXTENT_EARTHMODEL2_VISIR_RAPIDSCAN_FILL = {
    'earth_model': 2,
    'dataset_id': make_dataid(name='VIS006', resolution=3000),
    'is_full_disk': False,
    'is_rapid_scan': 1,
    'fill_disk': True,
    'expected_area_def': {
        'Area ID': 'msg_seviri_rss_3km',
        'Projection': {'a': '6378169000', 'b': '6356583800', 'h': '35785831',
                       'lon_0': '9.5', 'no_defs': 'None', 'proj': 'geos',
                       'type': 'crs', 'units': 'm', 'x_0': '0', 'y_0': '0'},
        'Number of columns': 3712,
        'Number of rows': 3712,
        'Area extent': (5567248.0742, 5570248.4773, -5570248.4773, -5567248.0742)

    }
}

TEST_AREA_EXTENT_EARTHMODEL2_HRV_RAPIDSCAN = {
    'earth_model': 2,
    'dataset_id': make_dataid(name='HRV', resolution=1000),
    'is_full_disk': False,
    'is_rapid_scan': 1,
    'fill_disk': False,
    'expected_area_def': {
        'Area ID': 'msg_seviri_rss_1km',
        'Projection': {'a': '6378169000', 'b': '6356583800', 'h': '35785831',
                       'lon_0': '0', 'no_defs': 'None', 'proj': 'geos',
                       'type': 'crs', 'units': 'm', 'x_0': '0', 'y_0': '0'},
        'Number of columns': 5568,
        'Number of rows': 8192,
        'Area extent': (5566247.718632221, 2626852.867305279, -2500.3358721733093, -5566247.718632221)
    }
}

TEST_AREA_EXTENT_EARTHMODEL2_HRV_RAPIDSCAN_FILL = {
    'earth_model': 2,
    'dataset_id': make_dataid(name='HRV', resolution=1000),
    'is_full_disk': False,
    'is_rapid_scan': 1,
    'fill_disk': True,
    'expected_area_def': {
        'Area ID': 'msg_seviri_rss_1km',
        'Projection': {'a': '6378169000', 'b': '6356583800', 'h': '35785831',
                       'lon_0': '0', 'no_defs': 'None', 'proj': 'geos',
                       'type': 'crs', 'units': 'm', 'x_0': '0', 'y_0': '0'},
        'Number of columns': 11136,
        'Number of rows': 11136,
        'Area extent': (5566247.718632221, 5571248.390376568, -5571248.390376568, -5566247.718632221)
    }
}

TEST_AREA_EXTENT_EARTHMODEL2_VISIR_ROI = {
    'earth_model': 2,
    'dataset_id': make_dataid(name='VIS006', resolution=3000),
    'is_full_disk': False,
    'is_rapid_scan': 0,
    'fill_disk': False,
    'expected_area_def': {
        'Area ID': 'msg_seviri_fes_3km',
        'Projection': {'a': '6378169000', 'b': '6356583800', 'h': '35785831',
                       'lon_0': '0', 'no_defs': 'None', 'proj': 'geos',
                       'type': 'crs', 'units': 'm', 'x_0': '0', 'y_0': '0'},
        'Number of columns': 2516,
        'Number of rows': 1829,
        'Area extent': (5336217.0304, 5156192.8405, -2212797.3348, -331544.5498)
    }
}

TEST_AREA_EXTENT_EARTHMODEL2_VISIR_ROI_FILL = {
    'earth_model': 2,
    'dataset_id': make_dataid(name='VIS006', resolution=3000),
    'is_full_disk': False,
    'is_rapid_scan': 0,
    'fill_disk': True,
    'expected_area_def': {
        'Area ID': 'msg_seviri_fes_3km',
        'Projection': {'a': '6378169000', 'b': '6356583800', 'h': '35785831',
                       'lon_0': '0', 'no_defs': 'None', 'proj': 'geos',
                       'type': 'crs', 'units': 'm', 'x_0': '0', 'y_0': '0'},
        'Number of columns': 3712,
        'Number of rows': 3712,
        'Area extent': (5567248.0742, 5570248.4773, -5570248.4773, -5567248.0742)
    }
}

TEST_AREA_EXTENT_EARTHMODEL2_HRV_ROI = {
    'earth_model': 2,
    'dataset_id': make_dataid(name='HRV', resolution=1000),
    'is_full_disk': False,
    'is_rapid_scan': 0,
    'fill_disk': False,
    'expected_area_def': {
        'Area ID': 'msg_seviri_fes_1km',
        'Projection': {'a': '6378169000', 'b': '6356583800', 'h': '35785831',
                       'lon_0': '0', 'no_defs': 'None', 'proj': 'geos',
                       'type': 'crs', 'units': 'm', 'x_0': '0', 'y_0': '0'},
        'Number of columns': 7548,
        'Number of rows': 5487,
        'Area extent': (5335216.684043407, 5157192.769944668, -2213797.381222248, -330544.4023013115)
    }
}

TEST_AREA_EXTENT_EARTHMODEL2_HRV_ROI_FILL = {
    'earth_model': 2,
    'dataset_id': make_dataid(name='HRV', resolution=1000),
    'is_full_disk': False,
    'is_rapid_scan': 0,
    'fill_disk': True,
    'expected_area_def': {
        'Area ID': 'msg_seviri_fes_1km',
        'Projection': {'a': '6378169000', 'b': '6356583800', 'h': '35785831',
                       'lon_0': '0', 'no_defs': 'None', 'proj': 'geos',
                       'type': 'crs', 'units': 'm', 'x_0': '0', 'y_0': '0'},
        'Number of columns': 11136,
        'Number of rows': 11136,
        'Area extent': (5566247.718632221, 5571248.390376568, -5571248.390376568, -5566247.718632221)
    }
}

TEST_IS_ROI_FULLDISK = {
    'is_full_disk': True,
    'is_rapid_scan': 0,
    'is_roi': False
}

TEST_IS_ROI_RAPIDSCAN = {
    'is_full_disk': False,
    'is_rapid_scan': 1,
    'is_roi': False
}

TEST_IS_ROI_ROI = {
    'is_full_disk': False,
    'is_rapid_scan': 0,
    'is_roi': True
}

TEST_CALIBRATION_MODE = {
    'earth_model': 1,
    'dataset_id': make_dataid(name='IR_108', calibration='radiance'),
    'is_full_disk': True,
    'is_rapid_scan': 0,
    'calibration': 'radiance',
    'CalSlope': [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 0.95, 0.96, 0.97],
    'CalOffset': [-1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0],
    'GSICSCalCoeff': [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 0.95, 0.96, 0.97],
    'GSICSOffsetCount': [-51.0, -51.0, -51.0, -51.0, -51.0, -51.0, -51.0, -51.0, -51.0, -51.0, -51.0, -51.0]
}

TEST_PADDER_RSS_ROI = {
    'img_bounds': {'south': [2], 'north': [4], 'east': [2], 'west': [3]},
    'is_full_disk': False,
    'dataset_id': make_dataid(name='VIS006'),
    'dataset': xr.DataArray(np.ones((3, 2)), dims=['y', 'x']).astype(np.float32),
    'final_shape': (5, 5),
    'expected_padded_data': xr.DataArray(np.array([[np.nan, np.nan, np.nan, np.nan, np.nan],
                                                   [np.nan, 1.0, 1.0, np.nan, np.nan],
                                                   [np.nan, 1.0, 1.0, np.nan, np.nan],
                                                   [np.nan, 1.0, 1.0, np.nan, np.nan],
                                                   [np.nan, np.nan, np.nan, np.nan, np.nan]]),
                                         dims=['y', 'x']).astype(np.float32)
}

TEST_PADDER_FES_HRV = {
    'img_bounds': {'south': [1, 4], 'north': [3, 5], 'east': [2, 3], 'west': [3, 4]},
    'is_full_disk': True,
    'dataset_id': make_dataid(name='HRV'),
    'dataset': xr.DataArray(np.ones((5, 2)), dims=['y', 'x']).astype(np.float32),
    'final_shape': (5, 5),
    'expected_padded_data': xr.DataArray(np.array([[np.nan, 1.0, 1.0, np.nan, np.nan],
                                                   [np.nan, 1.0, 1.0, np.nan, np.nan],
                                                   [np.nan, 1.0, 1.0, np.nan, np.nan],
                                                   [np.nan, np.nan, 1.0, 1.0, np.nan],
                                                   [np.nan, np.nan, 1.0, 1.0, np.nan]]),
                                         dims=['y', 'x']).astype(np.float32)

}


class TestNativeMSGFileHandler(unittest.TestCase):
    """Test the NativeMSGFileHandler."""

    def test_get_available_channels(self):
        """Test the derivation of the available channel list."""
        available_chs = get_available_channels(TEST1_HEADER_CHNLIST)
        trues = ['WV_062', 'WV_073', 'IR_108', 'VIS006', 'VIS008', 'IR_120']
        for bandname in AVAILABLE_CHANNELS:
            if bandname in trues:
                self.assertTrue(available_chs[bandname])
            else:
                self.assertFalse(available_chs[bandname])

        available_chs = get_available_channels(TEST2_HEADER_CHNLIST)
        trues = ['VIS006', 'VIS008', 'IR_039', 'WV_062', 'WV_073', 'IR_087', 'HRV']
        for bandname in AVAILABLE_CHANNELS:
            if bandname in trues:
                self.assertTrue(available_chs[bandname])
            else:
                self.assertFalse(available_chs[bandname])

        available_chs = get_available_channels(TEST3_HEADER_CHNLIST)
        for bandname in AVAILABLE_CHANNELS:
            self.assertTrue(available_chs[bandname])


class TestNativeMSGArea(unittest.TestCase):
    """Test NativeMSGFileHandler.get_area_extent.

    The expected results have been verified by manually
    inspecting the output of geoferenced imagery.
    """

    @staticmethod
    def create_test_header(earth_model, dataset_id, is_full_disk, is_rapid_scan):
        """Create mocked NativeMSGFileHandler.

        Contains sufficient attributes for NativeMSGFileHandler.get_area_extent to be able to execute.
        """
        if dataset_id['name'] == 'HRV':
            reference_grid = 'ReferenceGridHRV'
            column_dir_grid_step = 1.0001343488693237
            line_dir_grid_step = 1.0001343488693237
        else:
            reference_grid = 'ReferenceGridVIS_IR'
            column_dir_grid_step = 3.0004031658172607
            line_dir_grid_step = 3.0004031658172607

        if is_full_disk:
            north = 3712
            east = 1
            west = 3712
            south = 1
            n_visir_cols = 3712
            n_visir_lines = 3712
            n_hrv_cols = 11136
            n_hrv_lines = 11136
            ssp_lon = 0
        elif is_rapid_scan:
            north = 3712
            east = 1
            west = 3712
            south = 2321
            n_visir_cols = 3712
            n_visir_lines = 1392
            n_hrv_cols = 11136
            n_hrv_lines = 4176
            ssp_lon = 9.5
        else:
            north = 3574
            east = 78
            west = 2591
            south = 1746
            n_visir_cols = 2516
            n_visir_lines = north - south + 1
            n_hrv_cols = n_visir_cols * 3
            n_hrv_lines = n_visir_lines * 3
            ssp_lon = 0

        header = {
            '15_DATA_HEADER': {
                'ImageDescription': {
                    reference_grid: {
                        'ColumnDirGridStep': column_dir_grid_step,
                        'LineDirGridStep': line_dir_grid_step,
                        'GridOrigin': 2,  # south-east corner
                    },
                    'ProjectionDescription': {
                        'LongitudeOfSSP': ssp_lon
                    }
                },
                'GeometricProcessing': {
                    'EarthModel': {
                        'TypeOfEarthModel': earth_model,
                        'EquatorialRadius': 6378169.0,
                        'NorthPolarRadius': 6356583.800000001,
                        'SouthPolarRadius': 6356583.800000001,
                    }
                },
                'SatelliteStatus': {
                    'SatelliteDefinition': {
                        'SatelliteId': 324
                    }
                }
            },
            '15_SECONDARY_PRODUCT_HEADER': {
                'NorthLineSelectedRectangle': {'Value': north},
                'EastColumnSelectedRectangle': {'Value': east},
                'WestColumnSelectedRectangle': {'Value': west},
                'SouthLineSelectedRectangle': {'Value': south},
                'SelectedBandIDs': {'Value': 'xxxxxxxxxxxx'},
                'NumberColumnsVISIR': {'Value': n_visir_cols},
                'NumberLinesVISIR': {'Value': n_visir_lines},
                'NumberColumnsHRV': {'Value': n_hrv_cols},
                'NumberLinesHRV': {'Value': n_hrv_lines},
            }

        }

        return header

    @staticmethod
    def create_test_trailer(is_rapid_scan):
        """Create Test Trailer.

        Mocked Trailer with sufficient attributes for
        NativeMSGFileHandler.get_area_extent to be able to execute.
        """
        trailer = {
            '15TRAILER': {
                'ImageProductionStats': {
                    'ActualL15CoverageHRV': {
                        'UpperNorthLineActual': 11136,
                        'UpperWestColumnActual': 7533,
                        'UpperSouthLineActual': 8193,
                        'UpperEastColumnActual': 1966,
                        'LowerNorthLineActual': 8192,
                        'LowerWestColumnActual': 5568,
                        'LowerSouthLineActual': 1,
                        'LowerEastColumnActual': 1
                    },
                    'ActualScanningSummary': {
                        'ReducedScan': is_rapid_scan
                    }
                }
            }
        }

        return trailer

    def prepare_area_defs(self, test_dict):
        """Prepare calculated and expected area definitions for equal checking."""
        earth_model = test_dict['earth_model']
        dataset_id = test_dict['dataset_id']
        is_full_disk = test_dict['is_full_disk']
        is_rapid_scan = test_dict['is_rapid_scan']
        fill_disk = test_dict['fill_disk']
        header = self.create_test_header(earth_model, dataset_id, is_full_disk, is_rapid_scan)
        trailer = self.create_test_trailer(is_rapid_scan)
        expected_area_def = test_dict['expected_area_def']

        with mock.patch('satpy.readers.seviri_l1b_native.np.fromfile') as fromfile, \
                mock.patch('satpy.readers.seviri_l1b_native.recarray2dict') as recarray2dict, \
                mock.patch('satpy.readers.seviri_l1b_native.NativeMSGFileHandler._get_memmap') as _get_memmap, \
                mock.patch('satpy.readers.seviri_l1b_native.NativeMSGFileHandler._read_trailer'), \
                mock.patch(
                    'satpy.readers.seviri_l1b_native.NativeMSGFileHandler._has_archive_header'
                ) as _has_archive_header:
            _has_archive_header.return_value = True
            fromfile.return_value = header
            recarray2dict.side_effect = (lambda x: x)
            _get_memmap.return_value = np.arange(3)
            fh = NativeMSGFileHandler(None, {}, None)
            fh.fill_disk = fill_disk
            fh.header = header
            fh.trailer = trailer
            fh.image_boundaries = ImageBoundaries(header, trailer, fh.mda)
            calc_area_def = fh.get_area_def(dataset_id)

        return (calc_area_def, expected_area_def)

    # Earth model 1 tests
    def test_earthmodel1_visir_fulldisk(self):
        """Test the VISIR FES with the EarthModel 1."""
        calculated, expected = self.prepare_area_defs(
            TEST_AREA_EXTENT_EARTHMODEL1_VISIR_FULLDISK
        )
        np.testing.assert_allclose(np.array(calculated.area_extent),
                                   np.array(expected['Area extent']))
        self.assertEqual(calculated.width, expected['Number of columns'])
        self.assertEqual(calculated.height, expected['Number of rows'])
        self.assertEqual(calculated.area_id, expected['Area ID'])

    def test_earthmodel1_hrv_fulldisk(self):
        """Test the HRV FES with the EarthModel 1."""
        calculated, expected = self.prepare_area_defs(
            TEST_AREA_EXTENT_EARTHMODEL1_HRV_FULLDISK
        )
        np.testing.assert_allclose(np.array(calculated.defs[0].area_extent),
                                   np.array(expected['Area extent 0']))
        np.testing.assert_allclose(np.array(calculated.defs[1].area_extent),
                                   np.array(expected['Area extent 1']))

        self.assertEqual(calculated.width, expected['Number of columns'])
        self.assertEqual(calculated.height, expected['Number of rows'])
        self.assertEqual(calculated.defs[0].area_id, expected['Area ID'])
        self.assertEqual(calculated.defs[1].area_id, expected['Area ID'])

    def test_earthmodel1_hrv_fulldisk_fill(self):
        """Test the HRV FES padded to fulldisk with the EarthModel 1."""
        calculated, expected = self.prepare_area_defs(
            TEST_AREA_EXTENT_EARTHMODEL1_HRV_FULLDISK_FILL
        )
        np.testing.assert_allclose(np.array(calculated.area_extent),
                                   np.array(expected['Area extent']))
        self.assertEqual(calculated.width, expected['Number of columns'])
        self.assertEqual(calculated.height, expected['Number of rows'])
        self.assertEqual(calculated.area_id, expected['Area ID'])

    def test_earthmodel1_visir_rapidscan(self):
        """Test the VISIR RSS with the EarthModel 1."""
        calculated, expected = self.prepare_area_defs(
            TEST_AREA_EXTENT_EARTHMODEL1_VISIR_RAPIDSCAN
        )

        np.testing.assert_allclose(np.array(calculated.area_extent),
                                   np.array(expected['Area extent']))

        self.assertEqual(calculated.width, expected['Number of columns'])
        self.assertEqual(calculated.height, expected['Number of rows'])
        self.assertEqual(calculated.area_id, expected['Area ID'])

    def test_earthmodel1_visir_rapidscan_fill(self):
        """Test the VISIR RSS padded to fulldisk with the EarthModel 1."""
        calculated, expected = self.prepare_area_defs(
            TEST_AREA_EXTENT_EARTHMODEL1_VISIR_RAPIDSCAN_FILL
        )

        np.testing.assert_allclose(np.array(calculated.area_extent),
                                   np.array(expected['Area extent']))

        self.assertEqual(calculated.width, expected['Number of columns'])
        self.assertEqual(calculated.height, expected['Number of rows'])
        self.assertEqual(calculated.area_id, expected['Area ID'])

    def test_earthmodel1_hrv_rapidscan(self):
        """Test the HRV RSS with the EarthModel 1."""
        calculated, expected = self.prepare_area_defs(
            TEST_AREA_EXTENT_EARTHMODEL1_HRV_RAPIDSCAN
        )

        np.testing.assert_allclose(np.array(calculated.area_extent),
                                   np.array(expected['Area extent']))

        self.assertEqual(calculated.width, expected['Number of columns'])
        self.assertEqual(calculated.height, expected['Number of rows'])
        self.assertEqual(calculated.area_id, expected['Area ID'])

    def test_earthmodel1_hrv_rapidscan_fill(self):
        """Test the HRV RSS padded to fulldisk with the EarthModel 1."""
        calculated, expected = self.prepare_area_defs(
            TEST_AREA_EXTENT_EARTHMODEL1_HRV_RAPIDSCAN_FILL
        )

        np.testing.assert_allclose(np.array(calculated.area_extent),
                                   np.array(expected['Area extent']))

        self.assertEqual(calculated.width, expected['Number of columns'])
        self.assertEqual(calculated.height, expected['Number of rows'])
        self.assertEqual(calculated.area_id, expected['Area ID'])

    def test_earthmodel1_visir_roi(self):
        """Test the VISIR ROI with the EarthModel 1."""
        calculated, expected = self.prepare_area_defs(
            TEST_AREA_EXTENT_EARTHMODEL1_VISIR_ROI
        )
        np.testing.assert_allclose(np.array(calculated.area_extent),
                                   np.array(expected['Area extent']))
        self.assertEqual(calculated.width, expected['Number of columns'])
        self.assertEqual(calculated.height, expected['Number of rows'])
        self.assertEqual(calculated.area_id, expected['Area ID'])

    def test_earthmodel1_visir_roi_fill(self):
        """Test the VISIR ROI padded to fulldisk with the EarthModel 1."""
        calculated, expected = self.prepare_area_defs(
            TEST_AREA_EXTENT_EARTHMODEL1_VISIR_ROI_FILL
        )
        np.testing.assert_allclose(np.array(calculated.area_extent),
                                   np.array(expected['Area extent']))
        self.assertEqual(calculated.width, expected['Number of columns'])
        self.assertEqual(calculated.height, expected['Number of rows'])
        self.assertEqual(calculated.area_id, expected['Area ID'])

    def test_earthmodel1_hrv_roi(self):
        """Test the HRV ROI with the EarthModel 1."""
        calculated, expected = self.prepare_area_defs(
            TEST_AREA_EXTENT_EARTHMODEL1_HRV_ROI
        )
        np.testing.assert_allclose(np.array(calculated.area_extent),
                                   np.array(expected['Area extent']))
        self.assertEqual(calculated.width, expected['Number of columns'])
        self.assertEqual(calculated.height, expected['Number of rows'])
        self.assertEqual(calculated.area_id, expected['Area ID'])

    def test_earthmodel1_hrv_roi_fill(self):
        """Test the HRV ROI padded to fulldisk with the EarthModel 1."""
        calculated, expected = self.prepare_area_defs(
            TEST_AREA_EXTENT_EARTHMODEL1_HRV_ROI_FILL
        )
        np.testing.assert_allclose(np.array(calculated.area_extent),
                                   np.array(expected['Area extent']))
        self.assertEqual(calculated.width, expected['Number of columns'])
        self.assertEqual(calculated.height, expected['Number of rows'])
        self.assertEqual(calculated.area_id, expected['Area ID'])

    # Earth model 2 tests
    def test_earthmodel2_visir_fulldisk(self):
        """Test the VISIR FES with the EarthModel 2."""
        calculated, expected = self.prepare_area_defs(
            TEST_AREA_EXTENT_EARTHMODEL2_VISIR_FULLDISK
        )
        np.testing.assert_allclose(np.array(calculated.area_extent),
                                   np.array(expected['Area extent']))
        self.assertEqual(calculated.width, expected['Number of columns'])
        self.assertEqual(calculated.height, expected['Number of rows'])
        self.assertEqual(calculated.area_id, expected['Area ID'])

    def test_earthmodel2_hrv_fulldisk(self):
        """Test the HRV FES with the EarthModel 2."""
        calculated, expected = self.prepare_area_defs(
            TEST_AREA_EXTENT_EARTHMODEL2_HRV_FULLDISK
        )
        np.testing.assert_allclose(np.array(calculated.defs[0].area_extent),
                                   np.array(expected['Area extent 0']))
        np.testing.assert_allclose(np.array(calculated.defs[1].area_extent),
                                   np.array(expected['Area extent 1']))

        self.assertEqual(calculated.width, expected['Number of columns'])
        self.assertEqual(calculated.height, expected['Number of rows'])
        self.assertEqual(calculated.defs[0].area_id, expected['Area ID'])
        self.assertEqual(calculated.defs[1].area_id, expected['Area ID'])

    def test_earthmodel2_hrv_fulldisk_fill(self):
        """Test the HRV FES padded to fulldisk with the EarthModel 2."""
        calculated, expected = self.prepare_area_defs(
            TEST_AREA_EXTENT_EARTHMODEL2_HRV_FULLDISK_FILL
        )
        np.testing.assert_allclose(np.array(calculated.area_extent),
                                   np.array(expected['Area extent']))
        self.assertEqual(calculated.width, expected['Number of columns'])
        self.assertEqual(calculated.height, expected['Number of rows'])
        self.assertEqual(calculated.area_id, expected['Area ID'])

    def test_earthmodel2_visir_rapidscan(self):
        """Test the VISIR RSS with the EarthModel 2."""
        calculated, expected = self.prepare_area_defs(
            TEST_AREA_EXTENT_EARTHMODEL2_VISIR_RAPIDSCAN
        )
        np.testing.assert_allclose(np.array(calculated.area_extent),
                                   np.array(expected['Area extent']))

        self.assertEqual(calculated.width, expected['Number of columns'])
        self.assertEqual(calculated.height, expected['Number of rows'])
        self.assertEqual(calculated.area_id, expected['Area ID'])

    def test_earthmodel2_visir_rapidscan_fill(self):
        """Test the VISIR RSS padded to fulldisk with the EarthModel 2."""
        calculated, expected = self.prepare_area_defs(
            TEST_AREA_EXTENT_EARTHMODEL2_VISIR_RAPIDSCAN_FILL
        )
        np.testing.assert_allclose(np.array(calculated.area_extent),
                                   np.array(expected['Area extent']))

        self.assertEqual(calculated.width, expected['Number of columns'])
        self.assertEqual(calculated.height, expected['Number of rows'])
        self.assertEqual(calculated.area_id, expected['Area ID'])

    def test_earthmodel2_hrv_rapidscan(self):
        """Test the HRV RSS with the EarthModel 2."""
        calculated, expected = self.prepare_area_defs(
            TEST_AREA_EXTENT_EARTHMODEL2_HRV_RAPIDSCAN
        )
        np.testing.assert_allclose(np.array(calculated.area_extent),
                                   np.array(expected['Area extent']))

        self.assertEqual(calculated.width, expected['Number of columns'])
        self.assertEqual(calculated.height, expected['Number of rows'])
        self.assertEqual(calculated.area_id, expected['Area ID'])

    def test_earthmodel2_hrv_rapidscan_fill(self):
        """Test the HRV RSS padded to fulldisk with the EarthModel 2."""
        calculated, expected = self.prepare_area_defs(
            TEST_AREA_EXTENT_EARTHMODEL2_HRV_RAPIDSCAN_FILL
        )
        np.testing.assert_allclose(np.array(calculated.area_extent),
                                   np.array(expected['Area extent']))

        self.assertEqual(calculated.width, expected['Number of columns'])
        self.assertEqual(calculated.height, expected['Number of rows'])
        self.assertEqual(calculated.area_id, expected['Area ID'])

    def test_earthmodel2_visir_roi(self):
        """Test the VISIR ROI with the EarthModel 2."""
        calculated, expected = self.prepare_area_defs(
            TEST_AREA_EXTENT_EARTHMODEL2_VISIR_ROI
        )
        np.testing.assert_allclose(np.array(calculated.area_extent),
                                   np.array(expected['Area extent']))
        self.assertEqual(calculated.width, expected['Number of columns'])
        self.assertEqual(calculated.height, expected['Number of rows'])
        self.assertEqual(calculated.area_id, expected['Area ID'])

    def test_earthmodel2_visir_roi_fill(self):
        """Test the VISIR ROI padded to fulldisk with the EarthModel 2."""
        calculated, expected = self.prepare_area_defs(
            TEST_AREA_EXTENT_EARTHMODEL2_VISIR_ROI_FILL
        )
        np.testing.assert_allclose(np.array(calculated.area_extent),
                                   np.array(expected['Area extent']))
        self.assertEqual(calculated.width, expected['Number of columns'])
        self.assertEqual(calculated.height, expected['Number of rows'])
        self.assertEqual(calculated.area_id, expected['Area ID'])

    def test_earthmodel2_hrv_roi(self):
        """Test the HRV ROI with the EarthModel 2."""
        calculated, expected = self.prepare_area_defs(
            TEST_AREA_EXTENT_EARTHMODEL2_HRV_ROI
        )
        np.testing.assert_allclose(np.array(calculated.area_extent),
                                   np.array(expected['Area extent']))
        self.assertEqual(calculated.width, expected['Number of columns'])
        self.assertEqual(calculated.height, expected['Number of rows'])
        self.assertEqual(calculated.area_id, expected['Area ID'])

    def test_earthmodel2_hrv_roi_fill(self):
        """Test the HRV ROI padded to fulldisk with the EarthModel 2."""
        calculated, expected = self.prepare_area_defs(
            TEST_AREA_EXTENT_EARTHMODEL2_HRV_ROI_FILL
        )
        np.testing.assert_allclose(np.array(calculated.area_extent),
                                   np.array(expected['Area extent']))
        self.assertEqual(calculated.width, expected['Number of columns'])
        self.assertEqual(calculated.height, expected['Number of rows'])
        self.assertEqual(calculated.area_id, expected['Area ID'])

    # Test check for Region Of Interest (ROI) data
    def prepare_is_roi(self, test_dict):
        """Prepare calculated and expected check for region of interest data for equal checking."""
        earth_model = 2
        dataset_id = make_dataid(name='VIS006')
        is_full_disk = test_dict['is_full_disk']
        is_rapid_scan = test_dict['is_rapid_scan']
        header = self.create_test_header(earth_model, dataset_id, is_full_disk, is_rapid_scan)
        trailer = self.create_test_trailer(is_rapid_scan)
        expected_is_roi = test_dict['is_roi']

        with mock.patch('satpy.readers.seviri_l1b_native.np.fromfile') as fromfile, \
                mock.patch('satpy.readers.seviri_l1b_native.recarray2dict') as recarray2dict, \
                mock.patch('satpy.readers.seviri_l1b_native.NativeMSGFileHandler._get_memmap') as _get_memmap, \
                mock.patch('satpy.readers.seviri_l1b_native.NativeMSGFileHandler._read_trailer'), \
                mock.patch(
                    'satpy.readers.seviri_l1b_native.NativeMSGFileHandler._has_archive_header'
                ) as _has_archive_header:
            _has_archive_header.return_value = True
            fromfile.return_value = header
            recarray2dict.side_effect = (lambda x: x)
            _get_memmap.return_value = np.arange(3)
            fh = NativeMSGFileHandler(None, {}, None)
            fh.header = header
            fh.trailer = trailer
            calc_is_roi = fh.is_roi()

        return (calc_is_roi, expected_is_roi)

    def test_is_roi_fulldisk(self):
        """Test check for region of interest with FES data."""
        calculated, expected = self.prepare_is_roi(TEST_IS_ROI_FULLDISK)
        self.assertEqual(calculated, expected)

    def test_is_roi_rapidscan(self):
        """Test check for region of interest with RSS data."""
        calculated, expected = self.prepare_is_roi(TEST_IS_ROI_RAPIDSCAN)
        self.assertEqual(calculated, expected)

    def test_is_roi_roi(self):
        """Test check for region of interest with ROI data."""
        calculated, expected = self.prepare_is_roi(TEST_IS_ROI_ROI)
        self.assertEqual(calculated, expected)


TEST_HEADER_CALIB = {
    'RadiometricProcessing': {
        'Level15ImageCalibration': {
            'CalSlope': TestFileHandlerCalibrationBase.gains_nominal,
            'CalOffset': TestFileHandlerCalibrationBase.offsets_nominal,

        },
        'MPEFCalFeedback': {
            'GSICSCalCoeff': TestFileHandlerCalibrationBase.gains_gsics,
            'GSICSOffsetCount': TestFileHandlerCalibrationBase.offsets_gsics
        }
    },
    'ImageDescription': {
        'Level15ImageProduction': {
            'PlannedChanProcessing': TestFileHandlerCalibrationBase.radiance_types
        }
    },
}


class TestNativeMSGCalibration(TestFileHandlerCalibrationBase):
    """Unit tests for calibration."""

    @pytest.fixture(name='file_handler')
    def file_handler(self):
        """Create a mocked file handler."""
        header = {
            '15_DATA_HEADER': {
                'ImageAcquisition': {
                    'PlannedAcquisitionTime': {
                        'TrueRepeatCycleStart': self.scan_time
                    }
                }
            }
        }
        header['15_DATA_HEADER'].update(TEST_HEADER_CALIB)
        with mock.patch('satpy.readers.seviri_l1b_native.NativeMSGFileHandler.__init__',
                        return_value=None):
            fh = NativeMSGFileHandler()
            fh.header = header
            fh.platform_id = self.platform_id
            return fh

    @pytest.mark.parametrize(
        ('channel', 'calibration', 'calib_mode', 'use_ext_coefs'),
        [
            # VIS channel, internal coefficients
            ('VIS006', 'counts', 'NOMINAL', False),
            ('VIS006', 'radiance', 'NOMINAL', False),
            ('VIS006', 'radiance', 'GSICS', False),
            ('VIS006', 'reflectance', 'NOMINAL', False),
            # VIS channel, external coefficients (mode should have no effect)
            ('VIS006', 'radiance', 'GSICS', True),
            ('VIS006', 'reflectance', 'NOMINAL', True),
            # IR channel, internal coefficients
            ('IR_108', 'counts', 'NOMINAL', False),
            ('IR_108', 'radiance', 'NOMINAL', False),
            ('IR_108', 'radiance', 'GSICS', False),
            ('IR_108', 'brightness_temperature', 'NOMINAL', False),
            ('IR_108', 'brightness_temperature', 'GSICS', False),
            # IR channel, external coefficients (mode should have no effect)
            ('IR_108', 'radiance', 'NOMINAL', True),
            ('IR_108', 'brightness_temperature', 'GSICS', True),
            # HRV channel, internal coefficiens
            ('HRV', 'counts', 'NOMINAL', False),
            ('HRV', 'radiance', 'NOMINAL', False),
            ('HRV', 'radiance', 'GSICS', False),
            ('HRV', 'reflectance', 'NOMINAL', False),
            # HRV channel, external coefficients (mode should have no effect)
            ('HRV', 'radiance', 'GSICS', True),
            ('HRV', 'reflectance', 'NOMINAL', True),
        ]
    )
    def test_calibrate(
            self, file_handler, counts, channel, calibration, calib_mode,
            use_ext_coefs
    ):
        """Test the calibration."""
        external_coefs = self.external_coefs if use_ext_coefs else {}
        expected = self._get_expected(
            channel=channel,
            calibration=calibration,
            calib_mode=calib_mode,
            use_ext_coefs=use_ext_coefs
        )

        fh = file_handler
        fh.calib_mode = calib_mode
        fh.ext_calib_coefs = external_coefs

        dataset_id = make_dataid(name=channel, calibration=calibration)
        res = fh.calibrate(counts, dataset_id)
        xr.testing.assert_allclose(res, expected)


class TestNativeMSGDataset:
    """Tests for getting the dataset."""

    @pytest.fixture
    def file_handler(self):
        """Create a file handler for testing."""
        header = {
            '15_DATA_HEADER': {
                'SatelliteStatus': {
                    'SatelliteDefinition': {
                        'NominalLongitude': 0.0
                    },
                    'Orbit': {
                        'OrbitPolynomial': ORBIT_POLYNOMIALS
                    }
                },
                'ImageAcquisition': {
                    'PlannedAcquisitionTime': {
                        'TrueRepeatCycleStart': datetime(
                            2006, 1, 1, 12, 15, 9, 304888
                        )
                    }
                }
            },
        }
        header['15_DATA_HEADER'].update(TEST_HEADER_CALIB)
        mda = {
            'channel_list': ['VIS006', 'IR_108'],
            'number_of_lines': 4,
            'number_of_columns': 4,
            'is_full_disk': True,
            'platform_name': 'MSG-3',
            'offset_corrected': True,
            'projection_parameters': {
                'ssp_longitude': 0.0,
                'h': 35785831.0,
                'a': 6378169.0,
                'b': 6356583.8
            }
        }
        num_visir_cols = 5  # will be divided by 1.25 -> 4 columns
        visir_rec = [
            ('line_data', np.uint8, (num_visir_cols,)),
            ('acq_time', time_cds_short)
        ]
        vis006_line1 = (
            [1, 2, 3, 4, 5],  # line_data
            (1, 1000)  # acq_time (days, milliseconds)
        )
        vis006_line2 = ([6, 7, 8, 9, 10], (1, 2000))
        vis006_line3 = ([11, 12, 13, 14, 15], (1, 3000))
        vis006_line4 = ([16, 17, 18, 19, 20], (1, 4000))
        ir108_line1 = ([20, 19, 18, 17, 16], (1, 1000))
        ir108_line2 = ([15, 14, 13, 12, 11], (1, 2000))
        ir108_line3 = ([10, 9, 8, 7, 6], (1, 3000))
        ir108_line4 = ([5, 4, 3, 2, 1], (1, 4000))
        data = np.array(
            [[(vis006_line1,), (ir108_line1,)],
             [(vis006_line2,), (ir108_line2,)],
             [(vis006_line3,), (ir108_line3,)],
             [(vis006_line4,), (ir108_line4,)]],
            dtype=[('visir', visir_rec)]
        )
        with mock.patch('satpy.readers.seviri_l1b_native.NativeMSGFileHandler.__init__',
                        return_value=None):
            fh = NativeMSGFileHandler()
            fh.header = header
            fh.mda = mda
            fh.dask_array = da.from_array(data)
            fh.platform_id = 324
            fh.fill_disk = False
            fh.calib_mode = 'NOMINAL'
            fh.ext_calib_coefs = {}
            fh.include_raw_metadata = False
            fh.mda_max_array_size = 100
            return fh

    def test_get_dataset(self, file_handler):
        """Test getting the dataset."""
        dataset_id = make_dataid(
            name='VIS006',
            resolution=3000,
            calibration='counts'
        )
        dataset_info = {
            'units': '1',
            'wavelength': (1, 2, 3),
            'standard_name': 'counts'
        }
        dataset = file_handler.get_dataset(dataset_id, dataset_info)
        expected = xr.DataArray(
            np.array([[4., 32., 193., 5.],
                      [24., 112., 514., 266.],
                      [44., 192., 835., 527.],
                      [64., 273., 132., 788.]],
                     dtype=np.float32),
            dims=('y', 'x'),
            attrs={
                'orbital_parameters': {
                    'satellite_actual_longitude': -3.55117540817073,
                    'satellite_actual_latitude': -0.5711243456528018,
                    'satellite_actual_altitude': 35783296.150123544,
                    'satellite_nominal_longitude': 0.0,
                    'satellite_nominal_latitude': 0.0,
                    'projection_longitude': 0.0,
                    'projection_latitude': 0.0,
                    'projection_altitude': 35785831.0
                },
                'georef_offset_corrected': True,
                'platform_name': 'MSG-3',
                'sensor': 'seviri',
                'units': '1',
                'wavelength': (1, 2, 3),
                'standard_name': 'counts'
            }
        )
        expected['acq_time'] = ('y', [np.datetime64('1958-01-02 00:00:01'),
                                      np.datetime64('1958-01-02 00:00:02'),
                                      np.datetime64('1958-01-02 00:00:03'),
                                      np.datetime64('1958-01-02 00:00:04')])
        xr.testing.assert_equal(dataset, expected)
        assert 'raw_metadata' not in dataset.attrs
        assert_attrs_equal(dataset.attrs, expected.attrs, tolerance=1e-4)

    def test_get_dataset_with_raw_metadata(self, file_handler):
        """Test provision of raw metadata."""
        file_handler.include_raw_metadata = True
        dataset_id = make_dataid(
            name='VIS006',
            resolution=3000,
            calibration='counts'
        )
        dataset_info = {
            'units': '1',
            'wavelength': (1, 2, 3),
            'standard_name': 'counts'
        }
        res = file_handler.get_dataset(dataset_id, dataset_info)
        assert 'raw_metadata' in res.attrs

    def test_satpos_no_valid_orbit_polynomial(self, file_handler):
        """Test satellite position if there is no valid orbit polynomial."""
        file_handler.header['15_DATA_HEADER']['SatelliteStatus'][
            'Orbit']['OrbitPolynomial'] = ORBIT_POLYNOMIALS_INVALID
        dataset_id = make_dataid(
            name='VIS006',
            resolution=3000,
            calibration='counts'
        )
        dataset_info = {
            'units': '1',
            'wavelength': (1, 2, 3),
            'standard_name': 'counts'
        }
        res = file_handler.get_dataset(dataset_id, dataset_info)
        assert 'satellite_actual_longitude' not in res.attrs[
            'orbital_parameters']


class TestNativeMSGPadder(unittest.TestCase):
    """Test Padder of the native l1b seviri reader."""

    @staticmethod
    def prepare_padder(test_dict):
        """Initialize Padder and pad test data."""
        dataset_id = test_dict['dataset_id']
        img_bounds = test_dict['img_bounds']
        is_full_disk = test_dict['is_full_disk']
        dataset = test_dict['dataset']
        final_shape = test_dict['final_shape']
        expected_padded_data = test_dict['expected_padded_data']

        padder = Padder(dataset_id, img_bounds, is_full_disk)
        padder._final_shape = final_shape
        calc_padded_data = padder.pad_data(dataset)

        return (calc_padded_data, expected_padded_data)

    def test_padder_rss_roi(self):
        """Test padder for RSS and ROI data (applies to both VISIR and HRV)."""
        calculated, expected = self.prepare_padder(TEST_PADDER_RSS_ROI)
        np.testing.assert_array_equal(calculated, expected)

    def test_padder_fes_hrv(self):
        """Test padder for FES HRV data."""
        calculated, expected = self.prepare_padder(TEST_PADDER_FES_HRV)
        np.testing.assert_array_equal(calculated, expected)


class TestNativeMSGFilenames:
    """Test identification of Native format filenames."""

    @pytest.fixture
    def reader(self):
        """Return reader for SEVIRI Native format."""
        from satpy._config import config_search_paths
        from satpy.readers import load_reader

        reader_configs = config_search_paths(
            os.path.join("readers", "seviri_l1b_native.yaml"))
        reader = load_reader(reader_configs)
        return reader

    def test_file_pattern(self, reader):
        """Test file pattern matching."""
        filenames = [
            # Valid
            "MSG2-SEVI-MSG15-0100-NA-20080219094242.289000000Z",
            "MSG2-SEVI-MSG15-0201-NA-20080219094242.289000000Z",
            "MSG2-SEVI-MSG15-0301-NA-20080219094242.289000000Z-123456.nat",
            "MSG2-SEVI-MSG15-0401-NA-20080219094242.289000000Z-20201231181545-123456.nat",
            # Invalid
            "MSG2-SEVI-MSG15-010-NA-20080219094242.289000000Z",
        ]
        files = reader.select_files_from_pathnames(filenames)
        assert len(files) == 4


@pytest.mark.parametrize(
    'file_content,exp_header_size',
    [
        (b'FormatName                  : NATIVE', 450400),  # with ascii header
        (b'foobar', 445286),  # without ascii header
    ]
)
def test_header_type(file_content, exp_header_size):
    """Test identification of the file header type."""
    header = TestNativeMSGArea.create_test_header(
        dataset_id=make_dataid(name='VIS006', resolution=3000),
        earth_model=1,
        is_full_disk=True,
        is_rapid_scan=0
    )
    if file_content == b'foobar':
        header.pop('15_SECONDARY_PRODUCT_HEADER')
    with mock.patch('satpy.readers.seviri_l1b_native.np.fromfile') as fromfile, \
            mock.patch('satpy.readers.seviri_l1b_native.recarray2dict') as recarray2dict, \
            mock.patch('satpy.readers.seviri_l1b_native.NativeMSGFileHandler._get_memmap') as _get_memmap, \
            mock.patch('satpy.readers.seviri_l1b_native.NativeMSGFileHandler._read_trailer'), \
            mock.patch("builtins.open", mock.mock_open(read_data=file_content)):
        fromfile.return_value = header
        recarray2dict.side_effect = (lambda x: x)
        _get_memmap.return_value = np.arange(3)
        fh = NativeMSGFileHandler('myfile', {}, None)
        assert fh.header_type.itemsize == exp_header_size
        assert '15_SECONDARY_PRODUCT_HEADER' in fh.header



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2020 Satpy developers
#
# satpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# satpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with satpy.  If not, see <http://www.gnu.org/licenses/>.

"""EUMETSAT EPS-SG Visible/Infrared Imager (VII) readers base class."""

import logging
from datetime import datetime

from geotiepoints.viiinterpolator import tie_points_geo_interpolation, tie_points_interpolation

from satpy.readers.netcdf_utils import NetCDF4FileHandler
from satpy.readers.vii_utils import SCAN_ALT_TIE_POINTS, TIE_POINTS_FACTOR

logger = logging.getLogger(__name__)


class ViiNCBaseFileHandler(NetCDF4FileHandler):
    """Base reader class for VII products in netCDF format.

    Args:
        filename (str): File to read
        filename_info (dict): Dictionary with filename information
        filetype_info (dict): Dictionary with filetype information
        orthorect (bool): activates the orthorectification correction where available

    """

    def __init__(self, filename, filename_info, filetype_info, orthorect=False):
        """Prepare the class for dataset reading."""
        super().__init__(filename, filename_info, filetype_info, auto_maskandscale=True)

        # Saves the orthorectification flag
        self.orthorect = orthorect and filetype_info.get('orthorect', True)

        # Saves the interpolation flag
        self.interpolate = filetype_info.get('interpolate', True)

        try:
            longitude = self[filetype_info['cached_longitude']]
            latitude = self[filetype_info['cached_latitude']]

            if self.interpolate:
                self.longitude, self.latitude = self._perform_geo_interpolation(longitude, latitude)
            else:
                self.longitude, self.latitude = longitude, latitude

        except KeyError:
            logger.warning("Cached longitude and/or latitude datasets are not correctly defined in YAML file")
            self.longitude, self.latitude = None, None

    def _standardize_dims(self, variable):
        """Standardize dims to y, x."""
        if 'num_pixels' in variable.dims:
            variable = variable.rename({'num_pixels': 'x', 'num_lines': 'y'})
        if variable.dims[0] == 'x':
            variable = variable.transpose('y', 'x')
        return variable

    def get_dataset(self, dataset_id, dataset_info):
        """Get dataset using file_key in dataset_info."""
        var_key = dataset_info['file_key']
        logger.debug('Reading in file to get dataset with key %s.', var_key)

        if var_key == 'cached_longitude' and self.longitude is not None:
            variable = self.longitude.copy()
        elif var_key == 'cached_latitude' and self.latitude is not None:
            variable = self.latitude.copy()
        else:
            try:
                variable = self[var_key]
            except KeyError:
                logger.warning("Could not find key %s in NetCDF file, no valid Dataset created", var_key)
                return None

            # If the dataset is marked for interpolation, perform the interpolation from tie points to pixels
            if dataset_info.get('interpolate', False) and self.interpolate:
                variable = self._perform_interpolation(variable)

            # Perform the calibration if required
            if dataset_info.get('calibration') is not None:
                variable = self._perform_calibration(variable, dataset_info)

        # Perform the orthorectification if required
        if self.orthorect:
            orthorect_data_name = dataset_info.get('orthorect_data', None)
            if orthorect_data_name is not None:
                variable = self._perform_orthorectification(variable, orthorect_data_name)

        # If the dataset contains a longitude, change it to the interval [0., 360.) as natively in the product
        # since the unwrapping performed during the interpolation might have created values outside this range
        if dataset_info.get('standard_name', None) == 'longitude':
            variable %= 360.

        # Manage the attributes of the dataset
        variable.attrs.setdefault('units', None)

        variable.attrs.update(dataset_info)
        variable.attrs.update(self._get_global_attributes())
        variable = self._standardize_dims(variable)
        return variable

    @staticmethod
    def _perform_interpolation(variable):
        """Perform the interpolation from tie points to pixel points.

        Args:
            variable: xarray DataArray containing the dataset to interpolate.

        Returns:
            DataArray: array containing the interpolate values, all the original metadata
                       and the updated dimension names.

        """
        interpolated_values = tie_points_interpolation(
            [variable],
            SCAN_ALT_TIE_POINTS,
            TIE_POINTS_FACTOR
        )[0]
        new_variable = interpolated_values.rename(
            num_tie_points_act='num_pixels',
            num_tie_points_alt='num_lines'
        )
        new_variable.name = variable.name
        new_variable.attrs = variable.attrs
        return new_variable

    @staticmethod
    def _perform_geo_interpolation(longitude, latitude):
        """Perform the interpolation of geographic coodinates from tie points to pixel points.

        Args:
            longitude: xarray DataArray containing the longitude dataset to interpolate.
            latitude: xarray DataArray containing the longitude dataset to interpolate.

        Returns:
            tuple of arrays containing the interpolate values, all the original metadata
                    and the updated dimension names.

        """
        interpolated_longitude, interpolated_latitude = tie_points_geo_interpolation(
            longitude,
            latitude,
            SCAN_ALT_TIE_POINTS,
            TIE_POINTS_FACTOR
        )
        new_longitude = interpolated_longitude.rename(
            num_tie_points_act='num_pixels',
            num_tie_points_alt='num_lines'
        )
        new_longitude.name = longitude.name
        new_longitude.attrs = longitude.attrs
        new_latitude = interpolated_latitude.rename(
            num_tie_points_act='num_pixels',
            num_tie_points_alt='num_lines'
        )
        new_latitude.name = latitude.name
        new_latitude.attrs = latitude.attrs
        return new_longitude, new_latitude

    def _perform_orthorectification(self, variable, orthorect_data_name):
        """Perform the orthorectification."""
        raise NotImplementedError

    def _perform_calibration(self, variable, dataset_info):
        """Perform the calibration."""
        raise NotImplementedError

    def _get_global_attributes(self):
        """Create a dictionary of global attributes to be added to all datasets."""
        attributes = {
            'filename': self.filename,
            'start_time': self.start_time,
            'end_time': self.end_time,
            'spacecraft_name': self.spacecraft_name,
            'ssp_lon': self.ssp_lon,
            'sensor': self.sensor,
            'filename_start_time': self.filename_info['sensing_start_time'],
            'filename_end_time': self.filename_info['sensing_end_time'],
            'platform_name': self.spacecraft_name,
        }

        # Add a "quality_group" item to the dictionary with all the variables and attributes
        # which are found in the 'quality' group of the VII product
        quality_group = self['quality']
        quality_dict = {}
        for key in quality_group:
            # Add the values (as Numpy array) of each variable in the group where possible
            try:
                quality_dict[key] = quality_group[key].values
            except ValueError:
                quality_dict[key] = None
        # Add the attributes of the quality group
        quality_dict.update(quality_group.attrs)

        attributes['quality_group'] = quality_dict

        return attributes

    @property
    def start_time(self):
        """Get observation start time."""
        try:
            start_time = datetime.strptime(self['/attr/sensing_start_time_utc'], '%Y%m%d%H%M%S.%f')
        except ValueError:
            start_time = datetime.strptime(self['/attr/sensing_start_time_utc'], '%Y-%m-%d %H:%M:%S.%f')
        return start_time

    @property
    def end_time(self):
        """Get observation end time."""
        try:
            end_time = datetime.strptime(self['/attr/sensing_end_time_utc'], '%Y%m%d%H%M%S.%f')
        except ValueError:
            end_time = datetime.strptime(self['/attr/sensing_end_time_utc'], '%Y-%m-%d %H:%M:%S.%f')
        return end_time

    @property
    def spacecraft_name(self):
        """Return spacecraft name."""
        return self['/attr/spacecraft']

    @property
    def sensor(self):
        """Return sensor."""
        return self['/attr/instrument']

    @property
    def ssp_lon(self):
        """Return subsatellite point longitude."""
        # This parameter is not applicable to VII
        return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""test file handler baseclass."""

import unittest
from unittest import mock

import numpy as np
import pytest

from satpy.readers.file_handlers import BaseFileHandler
from satpy.tests.utils import FakeFileHandler


class TestBaseFileHandler(unittest.TestCase):
    """Test the BaseFileHandler."""

    def setUp(self):
        """Set up the test."""
        self._old_set = BaseFileHandler.__abstractmethods__
        BaseFileHandler._abstractmethods__ = set()
        self.fh = BaseFileHandler(
            'filename', {'filename_info': 'bla'}, 'filetype_info')

    def test_combine_times(self):
        """Combine times."""
        info1 = {'start_time': 1}
        info2 = {'start_time': 2}
        res = self.fh.combine_info([info1, info2])
        exp = {'start_time': 1}
        self.assertDictEqual(res, exp)
        res = self.fh.combine_info([info2, info1])
        exp = {'start_time': 1}
        self.assertDictEqual(res, exp)

        info1 = {'end_time': 1}
        info2 = {'end_time': 2}
        res = self.fh.combine_info([info1, info2])
        exp = {'end_time': 2}
        self.assertDictEqual(res, exp)
        res = self.fh.combine_info([info2, info1])
        exp = {'end_time': 2}
        self.assertDictEqual(res, exp)

    def test_combine_orbits(self):
        """Combine orbits."""
        info1 = {'start_orbit': 1}
        info2 = {'start_orbit': 2}
        res = self.fh.combine_info([info1, info2])
        exp = {'start_orbit': 1}
        self.assertDictEqual(res, exp)
        res = self.fh.combine_info([info2, info1])
        exp = {'start_orbit': 1}
        self.assertDictEqual(res, exp)

        info1 = {'end_orbit': 1}
        info2 = {'end_orbit': 2}
        res = self.fh.combine_info([info1, info2])
        exp = {'end_orbit': 2}
        self.assertDictEqual(res, exp)
        res = self.fh.combine_info([info2, info1])
        exp = {'end_orbit': 2}
        self.assertDictEqual(res, exp)

    @mock.patch('satpy.readers.file_handlers.SwathDefinition')
    def test_combine_area(self, sdef):
        """Combine area."""
        area1 = mock.MagicMock()
        area1.lons = np.arange(5)
        area1.lats = np.arange(5)
        area1.name = 'area1'

        area2 = mock.MagicMock()
        area2.lons = np.arange(5)
        area2.lats = np.arange(5)
        area2.name = 'area2'

        info1 = {'area': area1}
        info2 = {'area': area2}

        self.fh.combine_info([info1, info2])
        self.assertTupleEqual(sdef.call_args[1]['lons'].shape, (2, 5))
        self.assertTupleEqual(sdef.call_args[1]['lats'].shape, (2, 5))
        self.assertEqual(sdef.return_value.name, 'area1_area2')

    def test_combine_orbital_parameters(self):
        """Combine orbital parameters."""
        info1 = {'orbital_parameters': {'projection_longitude': 1,
                                        'projection_latitude': 1,
                                        'projection_altitude': 1,
                                        'satellite_nominal_longitude': 1,
                                        'satellite_nominal_latitude': 1,
                                        'satellite_actual_longitude': 1,
                                        'satellite_actual_latitude': 1,
                                        'satellite_actual_altitude': 1,
                                        'nadir_longitude': 1,
                                        'nadir_latitude': 1,
                                        'only_in_1': False}}
        info2 = {'orbital_parameters': {'projection_longitude': 2,
                                        'projection_latitude': 2,
                                        'projection_altitude': 2,
                                        'satellite_nominal_longitude': 2,
                                        'satellite_nominal_latitude': 2,
                                        'satellite_actual_longitude': 2,
                                        'satellite_actual_latitude': 2,
                                        'satellite_actual_altitude': 2,
                                        'nadir_longitude': 2,
                                        'nadir_latitude': 2,
                                        'only_in_2': True}}
        exp = {'orbital_parameters': {'projection_longitude': 1.5,
                                      'projection_latitude': 1.5,
                                      'projection_altitude': 1.5,
                                      'satellite_nominal_longitude': 1.5,
                                      'satellite_nominal_latitude': 1.5,
                                      'satellite_actual_longitude': 1.5,
                                      'satellite_actual_latitude': 1.5,
                                      'satellite_actual_altitude': 1.5,
                                      'nadir_longitude': 1.5,
                                      'nadir_latitude': 1.5,
                                      'only_in_1': False,
                                      'only_in_2': True}}
        res = self.fh.combine_info([info1, info2])
        self.assertDictEqual(res, exp)

        # Identity
        self.assertEqual(self.fh.combine_info([info1]), info1)

        # Empty
        self.fh.combine_info([{}])

    def test_file_is_kept_intact(self):
        """Test that the file object passed (string, path, or other) is kept intact."""
        open_file = mock.MagicMock()
        bfh = BaseFileHandler(open_file, {'filename_info': 'bla'}, 'filetype_info')
        assert bfh.filename == open_file

        from pathlib import Path
        filename = Path('/bla/bla.nc')
        bfh = BaseFileHandler(filename, {'filename_info': 'bla'}, 'filetype_info')
        assert isinstance(bfh.filename, Path)

    def tearDown(self):
        """Tear down the test."""
        BaseFileHandler.__abstractmethods__ = self._old_set


@pytest.mark.parametrize(
    ("file_type", "ds_file_type", "exp_result"),
    [
        ("fake1", "fake1", True),
        ("fake1", ["fake1"], True),
        ("fake1", ["fake1", "fake2"], True),
        ("fake1", ["fake2"], None),
        ("fake1", "fake2", None),
        ("fake1", "fake1_with_suffix", None),
    ]
)
def test_file_type_match(file_type, ds_file_type, exp_result):
    """Test that file type matching uses exactly equality."""
    fh = FakeFileHandler("some_file.txt", {}, {"file_type": file_type})
    assert fh.file_type_matches(ds_file_type) is exp_result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Test fixed bugs."""


from unittest.mock import patch

import dask.array as da
import numpy as np
from xarray import DataArray, Dataset

from satpy.tests.utils import make_dataid

abi_file_list = ['/data/OR_ABI-L1b-RadF-M3C01_G16_s20180722030423_e20180722041189_c20180722041235-118900_0.nc',
                 '/data/OR_ABI-L1b-RadF-M3C02_G16_s20180722030423_e20180722041190_c20180722041228-120000_0.nc',
                 '/data/OR_ABI-L1b-RadF-M3C03_G16_s20180722030423_e20180722041190_c20180722041237-119000_0.nc',
                 '/data/OR_ABI-L1b-RadF-M3C04_G16_s20180722030423_e20180722041189_c20180722041221.nc',
                 '/data/OR_ABI-L1b-RadF-M3C05_G16_s20180722030423_e20180722041190_c20180722041237-119101_0.nc',
                 '/data/OR_ABI-L1b-RadF-M3C06_G16_s20180722030423_e20180722041195_c20180722041227.nc',
                 '/data/OR_ABI-L1b-RadF-M3C07_G16_s20180722030423_e20180722041201_c20180722041238.nc',
                 '/data/OR_ABI-L1b-RadF-M3C08_G16_s20180722030423_e20180722041190_c20180722041238.nc',
                 '/data/OR_ABI-L1b-RadF-M3C09_G16_s20180722030423_e20180722041195_c20180722041256.nc',
                 '/data/OR_ABI-L1b-RadF-M3C10_G16_s20180722030423_e20180722041201_c20180722041250.nc',
                 '/data/OR_ABI-L1b-RadF-M3C11_G16_s20180722030423_e20180722041189_c20180722041254.nc',
                 '/data/OR_ABI-L1b-RadF-M3C12_G16_s20180722030423_e20180722041195_c20180722041256.nc',
                 '/data/OR_ABI-L1b-RadF-M3C13_G16_s20180722030423_e20180722041201_c20180722041259.nc',
                 '/data/OR_ABI-L1b-RadF-M3C14_G16_s20180722030423_e20180722041190_c20180722041258.nc',
                 '/data/OR_ABI-L1b-RadF-M3C15_G16_s20180722030423_e20180722041195_c20180722041259.nc',
                 '/data/OR_ABI-L1b-RadF-M3C16_G16_s20180722030423_e20180722041202_c20180722041259.nc']


def generate_fake_abi_xr_dataset(filename, chunks=None, **kwargs):
    """Create a fake xarray dataset for abi data.

    This is an incomplete copy of existing file structures.
    """
    dataset = Dataset(attrs={
        'time_coverage_start': '2018-03-13T20:30:42.3Z',
        'time_coverage_end': '2018-03-13T20:41:18.9Z',
    })

    projection = DataArray(
        [-214748364],
        attrs={
            'long_name': 'GOES-R ABI fixed grid projection',
            'grid_mapping_name': 'geostationary',
            'perspective_point_height': 35786023.0,
            'semi_major_axis': 6378137.0,
            'semi_minor_axis': 6356752.31414,
            'inverse_flattening': 298.2572221,
            'latitude_of_projection_origin': 0.0,
            'longitude_of_projection_origin': -75.0,
            'sweep_angle_axis': 'x'
        })
    dataset['goes_imager_projection'] = projection

    if 'C01' in filename or 'C03' in filename or 'C05' in filename:
        stop = 10847
        step = 2
        scale = 2.8e-05
        offset = 0.151858
    elif 'C02' in filename:
        stop = 21693
        step = 4
        scale = 1.4e-05
        offset = 0.151865
    else:
        stop = 5424
        step = 1
        scale = 5.6e-05
        offset = 0.151844

    y = DataArray(
        da.arange(0, stop, step),
        attrs={
            'scale_factor': -scale,
            'add_offset': offset,
            'units': 'rad',
            'axis': 'Y',
            'long_name': 'GOES fixed grid projection y-coordinate',
            'standard_name': 'projection_y_coordinate'
            },
        dims=['y'])

    dataset['y'] = y

    x = DataArray(
        da.arange(0, stop, step),
        attrs={
            'scale_factor': scale,
            'add_offset': -offset,
            'units': 'rad',
            'axis': 'X',
            'long_name': 'GOES fixed grid projection x-coordinate',
            'standard_name': 'projection_x_coordinate'
        },
        dims=['x'])

    dataset['x'] = x

    rad = DataArray(
        da.random.randint(0, 1025, size=[len(y), len(x)], dtype=np.int16, chunks=chunks),
        attrs={
            '_FillValue': np.array(1023),
            'long_name': 'ABI L1b Radiances',
            'standard_name': 'toa_outgoing_radiance_per_unit_wavelength',
            '_Unsigned': 'true',
            'sensor_band_bit_depth': 10,
            'valid_range': np.array([0, 1022], dtype=np.int16),
            'scale_factor': 0.8121064,
            'add_offset': -25.936647,
            'units': 'W m-2 sr-1 um-1',
            'resolution': 'y: 0.000028 rad x: 0.000028 rad',
            'grid_mapping': 'goes_imager_projection',
            'cell_methods': 't: point area: point'
        },
        dims=['y', 'x']
    )

    dataset['Rad'] = rad

    sublat = DataArray(0.0, attrs={
        'long_name': 'nominal satellite subpoint latitude (platform latitude)',
        'standard_name': 'latitude',
        '_FillValue': -999.0,
        'units': 'degrees_north'})
    dataset['nominal_satellite_subpoint_lat'] = sublat

    sublon = DataArray(-75.0, attrs={
        'long_name': 'nominal satellite subpoint longitude (platform longitude)',
        'standard_name': 'longitude',
        '_FillValue': -999.0,
        'units': 'degrees_east'})

    dataset['nominal_satellite_subpoint_lon'] = sublon

    satheight = DataArray(35786.023, attrs={
        'long_name': 'nominal satellite height above GRS 80 ellipsoid (platform altitude)',
        'standard_name': 'height_above_reference_ellipsoid',
        '_FillValue': -999.0,
        'units': 'km'})

    dataset['nominal_satellite_height'] = satheight

    yaw_flip_flag = DataArray(0, attrs={
        'long_name': 'Flag indicating the spacecraft is operating in yaw flip configuration',
        '_Unsigned': 'true',
        '_FillValue': np.array(-1),
        'valid_range': np.array([0, 1], dtype=np.int8),
        'units': '1',
        'flag_values': '0 1',
        'flag_meanings': 'false true'})

    dataset['yaw_flip_flag'] = yaw_flip_flag

    return dataset


@patch('xarray.open_dataset')
def test_1258(fake_open_dataset):
    """Save true_color from abi with radiance doesn't need two resamplings."""
    from satpy import Scene
    fake_open_dataset.side_effect = generate_fake_abi_xr_dataset

    scene = Scene(abi_file_list, reader='abi_l1b')
    scene.load(['true_color_nocorr', 'C04'], calibration='radiance')
    resampled_scene = scene.resample(scene.coarsest_area(), resampler='native')
    assert len(resampled_scene.keys()) == 2


@patch('xarray.open_dataset')
def test_1088(fake_open_dataset):
    """Check that copied arrays gets resampled."""
    from satpy import Scene
    fake_open_dataset.side_effect = generate_fake_abi_xr_dataset

    scene = Scene(abi_file_list, reader='abi_l1b')
    scene.load(['C04'], calibration='radiance')

    my_id = make_dataid(name='my_name', wavelength=(10, 11, 12))
    scene[my_id] = scene['C04'].copy()
    resampled = scene.resample('eurol')
    assert resampled[my_id].shape == (2048, 2560)


@patch('xarray.open_dataset')
def test_no_enums(fake_open_dataset):
    """Check that no enums are inserted in the resulting attrs."""
    from enum import Enum

    from satpy import Scene
    fake_open_dataset.side_effect = generate_fake_abi_xr_dataset

    scene = Scene(abi_file_list, reader='abi_l1b')
    scene.load(['C04'], calibration='radiance')
    for value in scene['C04'].attrs.values():
        assert not isinstance(value, Enum)
